// Given any Markov random field (MRF), with a finite number of discrete variables with finitely-many possible values,
// 1 - canonizes the energy,
// 2 - tests whether it belongs to Permutations(F_0),
// 3 - finds the representation in F_0,
// 4 - constructs the associated graph,
// 5 - and returns the solution.

// Makes use of the following algorithm :
// Given an interaction matrix between 2 variables, of an energy in the family F_0, permuted, finds the permutation Permut and the coefficients alpha_ij s.t. Permut(M) = \sum_ij alpha^ij W^ij

#ifndef FAMILIAL_PERMUTATIONS_H
#define FAMILIAL_PERMUTATIONS_H

// A few standard includes
#include <string>
#include <sstream>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#include <cmath>
#include <list>
#include <vector>
#include <algorithm>
#include <cassert>

// Personal libraries
#include "images.h"
#include "randomness.h"

// MRF - graph-cut by Kolmogorov and co
//#include "mrf.h"
//#include "ICM.h"
//#include "GCoptimization.h"
//#include "MaxProdBP.h"
//#include "graph.h"

namespace MRF_F0 {

// Interpretation of a variable x_i
// pair< Interpretation of x-i,  and for each neighbor, interpretation of the lines of the interaction matrix : pair< numbering of the label permutation chosen, node permutation > >
typedef     pair< vector<int>,  vector< pair<int, vector<int> > > > Interpretation;



/******************* Sorting methods ***********************/

// Decreasing order on the 2nd term
template <class A, class B>
bool sort_pair_2(pair<A,B> P, pair<A,B> Q){
  return (P.second > Q.second);
}

// Increasing order on the first term
template <class A, class B>
bool sort_pair_1(pair<A,B> P, pair<A,B> Q){
  return (P.first < Q.first);
}


// Increasing order on the 2nd term
template <class A, class B>
bool sort_pair_2_croissant(pair<A,B> P, pair<A,B> Q){
  return (P.second < Q.second);
}


/******************** Display *******************/

// Displays a vector in a terminal
template <class T>
void display_text(const vector<T>& V){

  typename vector<T>::const_iterator ita = V.begin(); 

  for ( ; ita != V.end(); ++ita)
    cout << (*ita) << "  ";
  cout << endl;
}


// to print a pair
template <class A, class B>
ostream& operator<<(ostream& F, const pair<A,B>& P){
  return F << "(" << P.first << ", " << P.second << ")";
}


// to print a list
template <class A>
ostream& operator<<(ostream& F, const list<A>& P){
  for (typename list<A>::const_iterator i = P.begin(); i != P.end(); ++i)
    F << (*i) << ", " ;
  return F;
}

// to print a vector<>
template <class A>
ostream& operator<<(ostream& F, const vector<A>& P){
  F << "[";
  for (typename vector<A>::const_iterator i = P.begin(); i != P.end(); ++i)
    F << (*i) << "; " ;
  F << "]";
  return F;
}

// to print a DoubleImage
template <class A>
ostream& operator<<(ostream& F, const RecImage<A>& M){
  int prec = 5;
  F << endl << "Size: " << M.nx << " x " << M.ny << endl;
  for (int i = 0; i<M.nx; i++){
    int nbd =  int( ceil( log(M.nx)/log(10) ) );
    char* b = new char[nbd+1];
    sprintf(b, "%*d", nbd, i);
    F << "L " << string(b) << " :  ";
    delete[] b;
    
    for (int j = 0; j<M.ny; j++){
      if (prec == -1){ // standard case
        F << "  " << M(i,j);
      } else { // case with alignment
        char* c = new char[prec+5];
        sprintf(c, "% -*.*g", prec+2, prec, double ( M(i,j) ) );
        F << "       " << string(c);
        delete[] c;
      }
    }
    F << endl;
  }
  return F;
}





/**************** Simple functions ****************/

// Swaps 2 elements of a table 
template <class T>
void swap_elements(vector<T>& V, int i, int j){
  T a = V[i];
  V[i] = V[j];
  V[j] = a;
}


// Returns the i-th bit of a (0 or 1)
inline int ith_bit(int i, int a){
  return (a / ( int(pow(2,i) + 0.1) ) ) % 2;
}


// factorielle...
int factorielle(int n){
  int a = 1;
  for (int i = 2; i <= n; i++)
    a *= i;
  return a;
}


/********************* Applying permutations ******************/

// Returns W = V o permut
template <class A>
vector<A> permute(const vector<A>& V, const vector<int>& permut){
  int n = V.size();
  assert(int(permut.size()) == n);
  vector<A> W(n);
  for (int i = 0; i<n; i++)
    W[i] = V[permut[i]];
  
  return W;
}


// Returns W = V o permut^-1
template <class A>
vector<A> permute_inv(const vector<A>& V, const vector<int>& permut){
  int n = V.size();
  assert(int(permut.size()) == n);
  vector<A> W(n);
  for (int i = 0; i<n; i++)
    W[permut[i]] = V[i];
  
  return W;
}

// Returns permut^-1
vector<int> inverse_permutation(const vector<int>& permut){
  int n = permut.size();
  vector<int> W(n);
  for (int i = 0; i<n; i++)
    W[permut[i]] = i;
  
  return W;
}



// Returns P = M o permut
template <class A>
RecImage<A> permute_x(const RecImage<A>& M, const vector<int>& permut){
  int n = M.nx;
  assert( int(permut.size()) == n);
  RecImage<A> P(n, M.ny);

  for (int i = 0; i<n; i++)
    for (int j = 0; j < M.ny ; j++)
      P(i,j) = M(permut[i],j);
	 
  return P;
}


// Returns P = M o permut^inv (permutation of one side of the matrix)
template <class A>
RecImage<A> permute_x_inv(const RecImage<A>& M, const vector<int>& permut){
  int n = M.nx;
  assert( int(permut.size()) == n);
  RecImage<A> P(n, M.ny);

  for (int i = 0; i<n; i++)
    for (int j = 0; j < M.ny ; j++)
      P(permut[i],j) = M(i,j);
	 
  return P;
}



// Idem over y instead of x
// Returns P = M o permut
template <class A>
RecImage<A> permute_y(const RecImage<A>& M, const vector<int>& permut){
  int n = M.ny;
  assert( int(permut.size()) == n);
  RecImage<A> P(M.nx, n);

  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j < n ; j++)
      P(i,j) = M(i, permut[j]);
	 
  return P;
}


// Returns P = M o permut^inv
template <class A>
RecImage<A> permute_y_inv(const RecImage<A>& M, const vector<int>& permut){
  int n = M.ny;
  assert( int(permut.size()) == n);
  RecImage<A> P(M.nx, n);

  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j < n ; j++)
      P(i, permut[j]) = M(i,j);
	 
  return P;
}

// Returns the composition of 2 permutations, i.e. A o B
// Beware the composition order !
vector<int> compose_permutations(const vector<int>& A, const vector<int>& B){
  int n = A.size();
  assert( int(B.size()) == n);
  vector<int> C(n);

  for (int i = 0; i<n; i++)
    C[i] = A[B[i]];

  return C;
}

// permutation that reverses the ordre
vector<int> reverse(int n){
  vector<int> V(n);
  for (int i = 0; i<n; i++)
    V[i] = n-i-1;
  return V;
}


// Reverses the indices
// Warning : modify the argument !
// V : subset of indices
// n : total nb of elements
// Ex: V = {2, 5, 6},  n = 10 : {0, 1, 2..., 9}
void reverse_index(vector<int>& V, int n){
  for (vector<int>::iterator i = V.begin(); i != V.end(); ++i)
    *i = n - (*i) - 1;
}

void reverse_index(vector< vector<int> >& V, int n){
  for (vector< vector<int> >::iterator i = V.begin(); i != V.end(); ++i)
    reverse_index(*i, n);
}



// Given a sequence of permutations p1, p2, p3, ..., p_last = id,  change it to p1, p1^-1 o p2, p2^-1 o p3, ..., p_beforelast^-1 in order to be able to apply them successively in loops
// Warning, applying the permutation p to the sequence V means computing V o p (and not p o V).
// Warning : modifies the argument...
void modified_sequence_of_permutations(vector< vector<int> >& liste_perm){
  int n = int(liste_perm.size());
  vector<int> p_v = liste_perm[0]; // p1 is not to be modified

  for (int i = 1; i<n; i++){
    vector<int> p_n = liste_perm[i];
    liste_perm[i] = compose_permutations(inverse_permutation(p_v) , p_n);
    p_v = p_n;
    p_n.clear();
  }
}

// Given a sequence of permutations p1, p2, p3, ..., p_last = id,  returns p1, p1^-1 o p2, p2^-1 o p3, ..., p_beforelast^-1 in order to be able to apply them successively in loops
// Warning, applying the permutation p to the sequence V means computing V o p (and not p o V).
vector< vector<int> >  sequence_of_permutations(const vector< vector<int> >& liste_perm){
  int n = int(liste_perm.size());
  vector< vector<int> >  Suite(n);

  Suite[0] = liste_perm[0]; // p1 is not to be modified

  for (int i = 1; i<n; i++)
    Suite[i] = compose_permutations(inverse_permutation(liste_perm[i-1]) , liste_perm[i]);

  return Suite;
}



// Takes indices (index) of matrix (M) and permutes them according to (permut)
void apply_permut_index_x(DoubleImage& M, const vector<int>& index, const vector<int>& permut ){
  int n = int(index.size());
  assert(int(permut.size()) == n);

  DoubleImage A(n, M.ny); // we permute lines...
  for (int i = 0; i<n; i++)
    for (int j = 0; j<A.ny; j++)
      A(i,j) = M(index[i],j);

  DoubleImage B = permute_x(A, permut);
  A.empty();

  for (int i = 0; i<n; i++)
    for (int j = 0; j<B.ny; j++)
      M(index[i],j) =  B(i,j);

  B.empty();
}


void apply_permut_index_y(DoubleImage& M, const vector<int>& index, const vector<int>& permut ){
  int n = int(index.size());
  assert(int(permut.size()) == n);

  DoubleImage A(M.nx, n); 
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<n; j++)
      A(i,j) = M(i,index[j]);

  DoubleImage B = permute_y(A, permut);
  A.empty();

  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<n; j++)
      M(i,index[j]) =  B(i,j);

  B.empty();
}



// Takes indices (index) of vector V and permutes them according to (permut)
template <class T>
void apply_permut_index(vector<T>& V, const vector<int>& index, const vector<int>& permut ){
  int n = int(index.size());
  assert(int(permut.size()) == n);

  vector<T> S (n); 
  for (int i = 0; i<n; i++)
    S[i] = V[index[i]];

  vector<T> B = permute(S, permut);
  S.clear();

  for (int i = 0; i<n; i++)
      V[index[i]] =  B[i];

  B.clear();
}



// Returns the permutation which makes nothing
vector<int> identity_permutation(int n){
  vector<int> Id(n);
  for (int i = 0; i<n; i++)
    Id[i] = i;
  return Id;
}


/******************* Random generations ****************/

// Generates n betas (positive or null!) and returns their 2^n combinations
vector<double> generate_betas_combined(int n){
  double* var = new double[n]; // the betas

  for (int i = 0; i<n; i++)
    var[i] = rand() * 1. /  RAND_MAX ; //generates a random double between 0 and 1
  
  int n_combi = int(pow(2,n) + 0.1);
  //    cout << "Nb variables : " << n << endl << "Nb combinations : " << n_combi << endl;
  
  // Combinations of values
  vector<double> val(n_combi);
  for (int j = 0; j<n_combi; j++){
    val[j] = 0;
    for (int i = 0; i<n; i++){
      int ibit = (j / ( int(pow(2,i) + 0.1) ) ) % 2; // i-th bit
      if (ibit == 0)
	val[j] += var[i];
      else
	val[j] -= var[i];
    }
  }
  
  delete[] var;
  return val;
}



// Generates a random permutation (uniform law over permutations)
vector<int> generate_permutation(int n){

  vector<int> P(n);
  for (int i = 0; i<n; i++)
    P[i] = i;

  for (int i = 1; i<n; i++){
    int j = rand() % (i+1);
    if (j != i)
      swap_elements(P, i, j);
  }

  return P;
}




// Generate a few alphas (positive or null)
DoubleImage generate_alphas(int m, int n){
  
  DoubleImage A(m,n);
  for (int i = 0; i<m; i++)
    for (int j = 0; j<n; j++)
      A(i,j) =  rand() * 1. /  RAND_MAX ; //generates a random double between 0 and 1
      //      A(i,j)= (rand() % 1000 ) / 1000. ; //generates a random double betweem 0 and 1, but in a more discrete set
  //  A(i,j)= (1 + (rand() % 100 )) / 100. ; // to avoid 0

  return A;
}




/**************** Decomposing and recomposing matrices ********************/

// Finds the alpha_ij corresponding to the matrix W^ij
// NB : there's a way to make this 4 times faster
// Warning, alphas are not guaranteed to be < 0.
double matrix_to_alpha(const DoubleImage& M, int i, int j){
  double a = 0;
  for (int im = 0; im<M.nx; im++)
    for (int jm = 0; jm<M.ny; jm++)
      if (ith_bit(i, im) + ith_bit(j,jm) == 1)
	a += M(im,jm);
      else
	a -= M(im,jm);

  a /=  (M.nx * M.ny);
  return a;
}


// Generates a matrix from the alphas (M = \sum_ij alpha_ij W^ij)
DoubleImage alphas_to_matrix(const DoubleImage& alphas){
  int nx = alphas.nx;
  int ny = alphas.ny;
  int tx = int( pow(2, nx) + 0.1 );
  int ty = int( pow(2, ny) + 0.1 );

  DoubleImage M(tx,ty);
  M.nul();

  for (int im = 0; im < tx; im++)
    for (int jm = 0; jm < ty; jm++)
      for (int i = 0; i < nx; i++)
	for (int j = 0; j < ny; j++){
	  int s = ( ith_bit(i, im) + ith_bit(j,jm) == 1  )?(1):(-1) ;
	  M(im,jm) += s * alphas(i,j);
	}

  //      M += alphas(i,j) * matrix_W(nx,ny,i,j);

  return M;

}



// Generates a matrix from the alphas and the interpretations of lines/columns (M = \sum_ij alpha_ij W^ij o Px^-1/Py^-1)
DoubleImage alphas_and_interp_to_matrix(const DoubleImage& alphas, const vector<int>& Px, const vector<int>& Py){
  int nx = alphas.nx;
  int ny = alphas.ny;
  int tx = int( pow(2, nx) + 0.1 );
  int ty = int( pow(2, ny) + 0.1 );

  DoubleImage M(tx,ty);
  M.nul();

  for (int im = 0; im < tx; im++)
    for (int jm = 0; jm < ty; jm++)
      for (int i = 0; i < nx; i++)
	for (int j = 0; j < ny; j++){
	  int s = ( ith_bit(i, im) + ith_bit(j,jm) == 1  )?(1):(-1) ;
	  M( Px[im], Py[jm] ) += s * alphas(i,j); // what a trick !
	}

  return M;
}






/************************* Belong-to tests ***********************/

// test whether an element belongs to a family of elements
template <class T>
bool belong_to(T a, const vector<T>& family){
  for (typename vector<T>::const_iterator i = family.begin(); i != family.end(); ++i){
    if ( *i == a )
      return true;
  }
  return false;
}


// test whether an element belongs to a subset of a set of family of elements
template <class T>
bool belong_to(T a, const vector< vector<T> >& grande_family, const vector<int>& sous_family){
  for (typename vector<int>::const_iterator i = sous_family.begin(); i != sous_family.end(); ++i){
    if ( belong_to(a, grande_family[ *i ] ) )
      return true;
  }
  return false;
}


// If M = \sum_ij alpha_ij W^ij + R,  returns R
DoubleImage compute_rest(const DoubleImage& M, const DoubleImage& alpha){
  return M - alphas_to_matrix(alpha);
}

// If M = \sum_ij alpha_ij W^ij + R, returns infinite_norm(R)
double distance_to_the_family(const DoubleImage& M, const DoubleImage& alpha){
  DoubleImage R = compute_rest(M, alpha);
  double Rn = norm_max(R);
  R.empty();
  return Rn;
}

bool is_in_the_family(const DoubleImage& M, const DoubleImage& alpha, double epsi = 0.000000001){
  for (int i = 0; i<alpha.nx; i++)
    for (int j = 0; j<alpha.ny; j++)
      if (alpha(i,j) < -epsi)
	return false;

  return (distance_to_the_family(M,alpha) < epsi);
}





/************************** Finding the betas from their +- sums *****************************/

// Given the liste of combinations +- of the betas, finds the betas (histogram method)
// liste_valeurs : the list of combination values ( M_000,s_2 )
// epsi : if |a-b| < epsi, then a = b (rounding mistakes)
// returns : the asolute values of the betas, sorted in decreasing order
// NB : normally all betas are supposed to be positive. If they were not, there would be no way to find their sign. Changing the sign of a beta is the same as inversing the bit of the associated node.
vector<double> find_betas_histo(const vector<double>& liste_valeurs, double epsi = 0.000000001){

  int n_combi = liste_valeurs.size();

  // Computes and sorts the differences between values
  vector<double> diff; // list of differences between 2 elements of val[]
  for (int j = 0; j<n_combi; j++)
    for (int k = j+1; k<n_combi; k++)
      diff.push_back( fabs(liste_valeurs[j] - liste_valeurs[k]) );
  
  sort(diff.begin(), diff.end());
    
  // Filling the histogram
  vector< pair<double, int> > histo; // list of values
  
  vector<double>::iterator it; // iterator on diff
  it = diff.begin();
  double vv = *it;
  histo.push_back( pair<double,int> (vv, 1) );
  
  for (++it ; it != diff.end(); ++it){
    if (fabs(*it - vv) < epsi)
      histo.back().second += 1;
    else {
      histo.push_back( pair<double,int> (*it, 1) );
      vv = *it;
    }
  }
  
  diff.clear();
  
  // Displays the histogram
  vector< pair<double,int> >::iterator itp; // iterator on histo
  //  for (itp = histo.begin(); itp != histo.end(); ++itp){
  //    cout << "Val : " << (*itp).first << "   ;   freq : " << (*itp).second << endl;
  //  }
  
  // sorts the histogram by frequency
  sort(histo.begin(), histo.end(), sort_pair_2<double,int>);
  
  
  // Displays the histogram once sorted
  // vector< pair<double,int> >::iterator itpt; // iterator on histo
  // for (itpt = histo.begin(); itpt != histo.end(); ++itpt){
  //   cout << (*itpt).first << "   ;   " << (*itpt).second << endl;
  // }
  
  // Most frequents elements
  itp = histo.begin(); 
  int fmax = (*itp).second;
  int fnouv;
  vector<double> betas;
  do {
    betas.push_back((*itp).first / 2);
    ++itp;
    fnouv = (*itp).second;
  } while (fnouv == fmax);
  
  histo.clear();
  sort(betas.begin(), betas.end());
  reverse(betas.begin(), betas.end());
  
  if (fnouv == 0)
    cout << "Frequency 2 : 0 !" << endl;
  else
    if (fmax != 2*fnouv)
      cout << "Frequency max vs 2nd : " << fmax << " vs " << fnouv << endl;
  
  // Displays results
  //  int i_nd = 0;
  //  cout << endl << "Results : freq = " << fmax << endl << endl;
  //  for (it = betas.begin(); it != betas.end(); ++it){
  //    i_nd++;
  //    cout << "Beta_" << i_nd << " : " << (*it) << endl;
  //  }
  
  

  if ( int(pow(2, betas.size()) + 0.1) != n_combi ){
    cout << "*************    Pb, not the good number of beta generators  ***************" << endl;
    //    cout << "List of values: "; display_text(liste_valeurs);
  }

  return betas;
}






// Returns the permutation to apply to the list of values "combines" in order that they be again in the order 000, 001, 010, 011, etc.
// betas : the betas which generate "combines"
// returns : vector<int> : permutation : perm[0010] = numbering of the line in "combines"
//          vector< vector<int> > : list of the sets of indices that create ambiguate permutations (in the case several values of "combines" are identical)
pair< vector<int>,  vector< vector<int> >   >   find_permutation(const vector<double>& betas, const vector<double>& combines, double epsi = 0.000000001){
  int nc = combines.size();
  int nb = betas.size();
  assert(int(pow(2,nb)+0.1) == nc);



  vector< pair<double,int> > combi2(nc);  // the combinations, generated in the right order (of decreasing betas)

  for (int j = 0; j<nc; j++){
    double a = 0;
    for (int i = 0; i<nb; i++){
      int ibit = (j / ( int(pow(2,i)+0.1) ) ) % 2; // i-th bit
      if (ibit == 0)
	a -= betas[i];
      else
	a += betas[i];
    }
    combi2[j] = pair<double,int>(a,j);
  }
  
  sort(combi2.begin(), combi2.end(), sort_pair_1<double,int>);



  
  vector< pair<double,int> > combi3(nc);  // initial combinations, numbered and to be sorted in the right order

  for (int j = 0; j<nc; j++)
    combi3[j] = pair<double,int>(combines[j],j);
  
  sort(combi3.begin(), combi3.end(), sort_pair_1<double,int>);




  // Correspondance
  
  vector<int> perm(nc);

  vector< pair<double,int> >::iterator it2 = combi2.begin();
  vector< pair<double,int> >::iterator it3 = combi3.begin();
  double vv = (*it2).first;
  int iv = (*it2).second; // binary index associated to vv

  //  cout << "Combi2: "; display_text(combi2);
  //  cout << "Combi3: "; display_text(combi3);

  vector< vector<int> > Ambig; // to store the sets of ambiguate indices (binary notation combi2)

  bool cycle_en_construction = false;
  for (int i = 0; i<nc; i++){

    double vn = (*it2).first;
    assert( fabs(vn - (*it3).first)  < epsi );
    perm[ (*it2).second ] = (*it3).second;

    if ( fabs(vv - vn) < epsi ) {
      if (i != 0){
	//	cout << "Warning, ambiguity in the variable values... (" << vv << " vs " << vn << "). Supplementary permutations to allow..." << endl;
	if (cycle_en_construction){
	  Ambig.back().push_back( (*it2).second );
	} else {
	  cycle_en_construction = true;
	  vector<int> deb(2);
	  deb[0] = iv;
	  deb[1] = (*it2).second;
	  Ambig.push_back(deb);
	}
      } 
    // otherwise i = 0 and c_e_c = already false
    } else {
      cycle_en_construction = false;
    }


    vv = vn;
    iv = (*it2).second;

    ++it2; ++it3;
  }
  combi2.clear();
  combi3.clear();



  //  cout << "Permut : "; display_text(perm);

  return pair< vector<int>, vector< vector<int> >   > (perm, Ambig);
}





// The same as before, except that we already know whether certain lines or columns are identical
// Returns the permutation to apply to the list of values "combines" in order that they be again in the order 000, 001, 010, 011, etc.
// betas : the betas generating "combines"
// Returns : vector<int> : permutation : perm[0010] = line number in "combines"
//          vector< vector<int> > : list of sets of indices creating ambiguous permutations (when several values in "combines" are identical)
pair< vector<int>,  vector< vector<int> >   >   find_permutations_and_sort(const vector<double>& betas, const vector<double>& combines, const vector< vector<int> >& lignes_identiques, double epsi = 0.000000001){
  int nc = combines.size();
  int nb = betas.size();
  assert(int(pow(2,nb)+0.1) == nc);
  assert(int(lignes_identiques.size()) == nc);


  vector< pair<double,int> > combi2(nc);  // the combinations, generated in the right order (of decreasing betas)

  for (int j = 0; j<nc; j++){
    double a = 0;
    for (int i = 0; i<nb; i++){
      int ibit = (j / ( int(pow(2,i)+0.1) ) ) % 2; // i-ith bit
      if (ibit == 0)
	a -= betas[i];
      else
	a += betas[i];
    }
    combi2[j] = pair<double,int>(a,j);
  }
  
  sort(combi2.begin(), combi2.end(), sort_pair_1<double,int>);



  
  vector< pair<double,int> > combi3(nc);  // initial combinations, numbered and to be sorted

  for (int j = 0; j<nc; j++)
    combi3[j] = pair<double,int>(combines[j],j);
  
  sort(combi3.begin(), combi3.end(), sort_pair_1<double,int>);




  // Correspondances 
  
  vector<int> perm(nc);

  vector< pair<double,int> >::iterator it2 = combi2.begin();
  vector< pair<double,int> >::iterator it3 = combi3.begin();
  double vv = (*it2).first;
  int iv2 = (*it2).second; // binary index associated to vv
  int iv3 = (*it3).second; // index in the permuted matrix

  //  cout << "Combi2: "; display_text(combi2);
  //  cout << "Combi3: "; display_text(combi3);

  vector< vector<int> > Ambig; // to store sets of anbiguous indices (binary notation combi2)

  bool cycle_en_construction = false;
  for (int i = 0; i<nc; i++){

    double vn = (*it2).first;
    assert( fabs(vn - (*it3).first)  < epsi );
    perm[ (*it2).second ] = (*it3).second;

    if ( fabs(vv - vn) < epsi ) {
      if (i != 0){
	if (cycle_en_construction){
	  if ( ! belong_to((*it3).second, lignes_identiques, Ambig.back() ) )
	    Ambig.back().push_back( (*it2).second );
	} else {
	  if ( ! belong_to((*it3).second, lignes_identiques[iv3]) ){
	    cycle_en_construction = true;
	    vector<int> deb(2);
	    deb[0] = iv2;
	    deb[1] = (*it2).second;
	    Ambig.push_back(deb);
	  }
	}
      } 
    // otherwise i = 0 and c_e_c = already false
    } else {
      cycle_en_construction = false;
    }


    vv = vn;
    iv2 = (*it2).second;
    iv3 = (*it3).second;

    ++it2; ++it3;
  }
  combi2.clear();
  combi3.clear();

  //  cout << "Permut : "; display_text(perm);

  return pair< vector<int>, vector< vector<int> >   > (perm, Ambig);
}





/*************** Finding extreme/particular elements in matrices/vectors *****************/

// Returns the coordinates of the biggest element in matrix A
template <class T>
pair<int,int> find_biggest(const RecImage<T>& A){
  T v = A(0,0);
  pair<int,int> P(0,0);
  for (int i = 0; i < A.nx; i++)
    for (int j = 0; j < A.ny; j++){
      T w = A(i,j);
      if (w>v){
	v = w;
	P = pair<int,int>(i,j);
      }
    }
  
  return P;
}


// Returns the coordinates of the biggest elementS in matrix A
template <class T>
list< pair<int,int> > find_biggests(const RecImage<T>& A, double epsi = 0.00000001){
  T v = A(0,0);
  pair<int,int> P(0,0);

  list< pair<int,int> > L; L.push_back(P);

  for (int i = 0; i < A.nx; i++)
    for (int j = 0; j < A.ny; j++){
      T w = A(i,j);
      if ( fabs(v-w) < epsi ){
	L.push_back( make_pair(i,j) );
      } else 
	if (w>v){
	  v = w;
	  P = pair<int,int>(i,j);
	  L.clear();
	  L.push_back(P);
	} 
    }
  
  return L;
}




// Returns the coordinates of the smallest element in vector A (argmin A)
template <class T>
int find_smallest(const vector<T>& A){
  T vm = A[0];
  int im = 0;
  for (int i = 0; i < int(A.size()); i++){
    if (A[i] < vm){
      vm = A[i];
      im = i;
    }
  }
  return im;
}


// Returns the coordinates of the biggest element in vector A (argmax A)
template <class T>
int find_biggest(const vector<T>& A){
  T vm = A[0];
  int im = 0;
  for (int i = 0; i < int(A.size()); i++){
    if (A[i] > vm){
      vm = A[i];
      im = i;
    }
  }
  return im;
}



// Returns the index of first element true
// If none, returns -1
int find_min_not_used(const vector<bool>& G){

  int i = 0;
  bool trouve = false;
  do {
    if ( !(G[i]) )
      trouve = true;
    else
      i++;
  } while ( (!trouve) && (i< int(G.size())) );

  if (trouve)
    return i;
  else
    return -1;
}


// Returns the first occurrence of a given value, not used yet, up to precision epsilon
int find_val_not_used(double v, const vector<double>& val, const vector<bool>& utilise, double epsi = 0.000000001){
  assert(val.size() == utilise.size());
  int i = 0;
  bool trouve = false;
  do {
    if (   ( !(utilise[i]) )   &&   ( fabs(val[i] - v) < epsi )   )
      trouve = true;
    else
      i++;
  } while ( (!trouve) && (i< int(val.size())) );

  if (trouve)
    return i;
  else
    return -1;
}


// Searches for a particular value (reached exactly once) and substitutes it
template <class T>
void find_unique_and_substitute(vector<T>& V, T val, T newval){
  for (int i = 0; i< int(V.size()); i++)
    if (V[i] == val){
      V[i] = newval;
      break;
    } 
}



// Searches for a value and returns the index of its first occurrence
// Returns -1 if not found
template <class T>
int find_value(const vector<T>& V, T val){
  for (int i = 0; i< int(V.size()); i++)
    if (V[i] == val){
      return i;
      break;
    } 
  return -1;
}


// Finds a value amongst first elements of pairs, and returns the index of first occurrence
// Returns -1 if not found
template <class T, class U>
int find_value_1(const vector< pair<T,U> >& V, T val){
  for (int i = 0; i< int(V.size()); i++)
    if (V[i].first == val){
      return i;
      break;
    } 
  return -1;
}

// Extracts a line or column from a matrix
// extract_x : vector in direction x, ie with M.nx terms
template <class T>
vector<T> extract_x(const RecImage<T>& M, int j){
  vector<T> V(M.nx);
  for (int i = 0; i<M.nx; i++)
    V[i] = M(i,j);

  return V;
}

template <class T>
vector<T> extract_y(const RecImage<T>& M, int i){
  vector<T> V(M.ny);
  for (int j = 0; j<M.ny; j++)
    V[j] = M(i,j);

  return V;
}




/**************************** Sorting matrix *************************/

// Re-sorts a matrix so that the sums of coefficients of lines and columns are decreasing
DoubleImage sort_decreasing_sums(const DoubleImage& alphas_init){
  vector< pair<double,int> > alphas_sum(alphas_init.ny);
  for (int i = 0; i<alphas_init.ny; i++){
    double aa = 0;
    for (int j = 0; j<alphas_init.nx; j++)
      aa += alphas_init(j,i);
    alphas_sum[i] = pair<double,int>(aa, i);
  }
  sort(alphas_sum.begin(), alphas_sum.end(), sort_pair_1<double,int>);
  reverse(alphas_sum.begin(), alphas_sum.end());
  vector<int> permay(alphas_init.ny);
  for (int i = 0; i<alphas_init.ny; i++)
    permay[i] = alphas_sum[i].second;
  DoubleImage alphas_init_ord = permute_y(alphas_init, permay);
  permay.clear();
  alphas_sum.clear();
  
  vector< pair<double,int> > alphas_sum2(alphas_init.nx);
  for (int i = 0; i<alphas_init.nx; i++){
    double aa = 0;
    for (int j = 0; j<alphas_init.ny; j++)
      aa += alphas_init(i,j);
    alphas_sum2[i] = pair<double,int>(aa, i);
  }
  sort(alphas_sum2.begin(), alphas_sum2.end(), sort_pair_1<double,int>);
  reverse(alphas_sum2.begin(), alphas_sum2.end());
  vector<int> permax(alphas_init.nx);
  for (int i = 0; i<alphas_init.nx; i++)
    permax[i] = alphas_sum2[i].second;
  //  cout << "Somme alphas: "; display_text(alphas_sum);
  alphas_sum2.clear();
  
  alphas_init_ord = permute_x(alphas_init_ord, permax);
  permax.clear();
  
  return alphas_init_ord;
}


/******************* Alternative algorithm to find the betas from their +- sums ***************************/

// Other way to find the absolute values of betas from the list of their signed sums.
// Is based on the graph whose nodes = values of sum \sum_i +- beta_i, and with edges  = + 2 beta_k 
// Returns an empty vector in case of pb.
vector<double> find_betas_graph(const vector<double>& liste_val, double epsi = 0.000000001){

  bool not_a_good_vector = false;

  vector<double> Val_ord(liste_val);
  sort(Val_ord.begin(), Val_ord.end()); // increasing order
  int N = Val_ord.size(); // N = 2^n : nb of values
  int n = int( log(N) / log(2) + 0.1); // nb of generators btas // + 0.1 because int(double) is not always right
  assert( int(pow(2,n)+0.1) == N);

  vector<bool> utilise(N, false); // to know if values have already been taken
  utilise[0] = true; // the root is the smallest sum

  vector<double> betas(n);
  vector<int> tree(N/2); // list of elements in the tree
                        // tree of (indices of) values, with edges  + 2 beta_k 
                        // N/2 because no need to build the end of the tree
  int nb_nodes = 0; // current number of nodes in the tree

  // A each step of the loop, we add a beta_k and its whole branch (as a partial ordered set)
  for (int nb_already_found = 0; nb_already_found<n; nb_already_found++){ 

    int a = find_min_not_used(utilise); // first other smallest sum
    assert(a != -1); // if we already explored the whole graph, we have a problem

    utilise[a] = true;
    double nouvbeta =  Val_ord[a] - Val_ord[0]; // new beta
    betas[nb_already_found] = nouvbeta;
    tree[nb_nodes] = a; // new leaf
    int nbnv = nb_nodes; // since we should not add branches to the new leaves we're going to add
    nb_nodes++;

    // generates the new branch
    // To each node already existing, add the branch +beta_k
    // It is sufficient to deal with branches with increasing +beta_k
    
    // In each step of the loop, we update a node by adding a new branche (of a single element, +beta_k)
    // (last loop step is useless : no need of the branch, thus we avoid building half of the tree, for nothing)
    if (nb_already_found != (n-1))
      for (int i = 0; i<nbnv; i++){ // should not deal with the new leaf
	assert(utilise[tree[i]]); // each node has already been used...
	int i_plus = find_val_not_used( Val_ord[ tree[i] ] + nouvbeta, Val_ord, utilise, epsi);
	not_a_good_vector =  (i_plus == -1);
	if (not_a_good_vector) 
	  break;

	utilise[i_plus] = true;
	tree[nb_nodes] = i_plus; // new branch
	nb_nodes++;
      }

    if (not_a_good_vector)
      break;
  }

  // Let's clean
  utilise.clear();
  tree.clear();
  
  if (not_a_good_vector){
    betas.clear();
  } else {

    // Update and checks
    double v = 0;
    for (int i = 0; i<n; i++){
      betas[i] /= 2; // +2 betas_k actually...
      v += betas[i];
    }

    double diff = fabs(v - Val_ord.back()); // sum of betas_i
    if (diff > epsi){
      betas.clear();
      return betas;
    }

    //  cout << "Betas : "; display_text(betas);

    // Sort betas in decreasing order
    sort(betas.begin(), betas.end());
    reverse(betas.begin(), betas.end());
  }
  
  // End of cleaning
  Val_ord.clear();

  return betas;

}



/*********** Norms, distances *************/

// infinite norm between 2 vectors
// T : int, double, float... or else you need to add  .norm()
template <class T>
double difference_max(const vector<T>& A, const vector<T>& B){
  double x = 0;
  int n = A.size();
  assert(n == int(B.size()));
  for (int i = 0; i<n; i++){
    double dx = fabs(A[i] - B[i]);
    if (dx > x)
      x = dx;
  }
  return x;
}


// Checks strict equality of 2 vectors (element by element)
template <class T>
bool is_strictly_equal(const vector<T>& A, const vector<T>& B){
  if (A.size() != B.size())
    return false;

  int n = int(A.size());
  for (int i = 0; i<n; i++){
    if (A[i] != B[i])
      return false;
  }

  return true;
}

// Checks approximate equality of 2 vectors (element par element)
template <class T>
bool is_equal(const vector<T>& A, const vector<T>& B, double epsi = 0.000000001){
  if (A.size() != B.size())
    return false;

  int n = int(A.size());
  for (int i = 0; i<n; i++){
    if ( fabs(A[i] - B[i]) > epsi )
      return false;
  }

  return true;
}


/***************** List of all permutations ********************/

// Generates all permutations, for all sets [1, 2, 3 ... n] for which lengths[n] = true
// output[n][i] = ith permutation with n elements
// last permutation = identity
vector< vector< vector<int> > > generate_all_permutations(const vector<bool>& lengths){
  int n_max = -1; // biggest number of elements to permute
  for (int i = 0; i<int(lengths.size()); i++)
    if (lengths[i])
      n_max = i;

  vector< vector< vector<int> > > Generated(n_max+1);

  int nb_perm = 1;
  vector< vector<int> > P_v(1); // old permut
  P_v[0] = vector<int>(1,0); // unique permutation of one element : {1}

  for (int i = 2; i<= n_max; i++){
    nb_perm *= i; // i!
    vector< vector<int> > P_i (nb_perm);
    int nk = 0; // index of the permutation within the set of permut with i elements
    for (int k = 0; k<i; k++){ // we end with element k
      for (int nk2 = 0; nk2<int(P_v.size()); nk2++){ // we go through all permutations with i-1 elements
	P_i[nk] = P_v[nk2];
	if (k != i-1)
	  find_unique_and_substitute(P_i[nk], k, i-1);
	P_i[nk].push_back(k);
	nk++;
      }
    }

    P_v = P_i;
    if (lengths[i])
      Generated[i] = P_i;
    P_i.clear();
  }		     

  P_v.clear();

  return Generated;

}



/*********************** To represent permutations of indices of an energy **********************/

// simple encapsulator
struct AlphaPermut {
  DoubleImage* alpha;
  vector<int>* px;
  vector<int>* py;

  AlphaPermut(){
    alpha = NULL; px = NULL; py = NULL;
  }

  AlphaPermut(DoubleImage* A, vector<int>* Px, vector<int>* Py){
    alpha = A; px = Px; py = Py;
  }

  // simple copy of pointers
  AlphaPermut(DoubleImage& A, vector<int>& Px, vector<int>& Py) : alpha(&A), px(&Px), py(&Py) {;}

  // creation of new matrices and pointers
  // boolean b is there just to distinguish the 2 constructors
  AlphaPermut(const DoubleImage& A, const vector<int>& Px, const vector<int>& Py,  bool b){
    b = true;

    alpha = new DoubleImage;
    copy(A, *alpha);
    px = new vector<int>;
    *px = Px;
    py = new vector<int>;
    *py = Py;
  }

  void empty(){
    if (alpha != NULL)
      alpha->empty();
    if (px != NULL)
      px->clear();
    if (py != NULL)
      py->clear();

    delete alpha; delete px; delete py;
    alpha = NULL; px = NULL; py = NULL;
  }


  // returns the good side of the interaction matrix (px or py) between x_i and the j-th neighbord of x_i (x_neighb_j, not x_j!), i.e. the side that stands for x_i.
  vector<int>*& p(const vector< vector< pair<int, int> > >& Neighbors, int i, int j){
	if ( Neighbors[i][j].first < i )
	  return py;
	else
	  return px;
  }

  vector<int>*& p(bool do_you_want_x_interrogation_mark){
	if ( do_you_want_x_interrogation_mark )
	  return px;
	else
	  return py;
  }

};

// Gives the permutation of nelem elements, whose coordinates are in coord, given the sets of elements to be permuted "index" as well as the permutations Perm of any length n.
vector<int> permutation_coordonnee(int nelem, const vector<int>& coord, const vector< vector<int> >& index, const vector< vector < vector<int> > >& Perm){
  int nc = coord.size();
  assert(nc == int( index.size() ) );

  vector<int> N = identity_permutation(nelem);
  
  for (int c = 0; c<nc; c++)
    if (coord[c] != 0)
      apply_permut_index(N, index[c], Perm[ index[c].size() ][coord[c]-1]);

  return N;
}





/******************************** Finding permutations ******************************/


// Given a matrix M, returns all permutations of this one for which the permuted M belongs to the family, as well as the corresponding alpha
// So : a list of pairs {alpha, permut}  s.t. M o permut = matrix_generated_by(alpha)
// Empty list if not a good matrix
// Symmetrization was missing (2 values max...) : now it's added
vector< AlphaPermut > find_alpha_and_permutation(const DoubleImage& M, bool check_similar_lines = false, bool verbose_multi = true, double epsi = 0.000000001){

  vector< AlphaPermut > Alpes;

  /******** Let's check first if there's no identical lines/columns ********/
  vector< vector<int> > similar_columns;
  if (check_similar_lines){
    similar_columns.resize(M.ny);
    for (int i = 0; i<M.ny; i++)
      for (int j = i+1; j<M.ny; j++){ // more efficient is possible...
	
	bool b = true;
	for (int k = 0; k<M.nx; k++)
	  if ( fabs(M(k,i) - M(k,j)) > epsi ){
	    b = false;
	    break;
	  }
	
	if (b){
	  similar_columns[i].push_back(j);
	  similar_columns[j].push_back(i);
	}
      }
  }

  vector< vector<int> > similar_lines;
  if (check_similar_lines){
    similar_lines.resize(M.nx);
    for (int i = 0; i<M.nx; i++)
      for (int j = i+1; j<M.nx; j++){ // more efficient is possible...
	
	bool b = true;
	for (int k = 0; k<M.ny; k++)
	  if ( fabs(M(i,k) - M(j,k)) > epsi){
	    b = false;
	    break;
	  }
	
	if (b){
	  similar_lines[i].push_back(j);
	  similar_lines[j].push_back(i);
	}
      }
  }

  if (check_similar_lines){
    int nb_s = 0;
    for (vector< vector<int> >::const_iterator i = similar_lines.begin(); i != similar_lines.end(); ++i)
      if ( (*i).size() != 0)
	nb_s ++;
    for (vector< vector<int> >::const_iterator i = similar_columns.begin(); i != similar_columns.end(); ++i)
      if ( (*i).size() != 0)
	nb_s ++;
    
    if (nb_s != 0)
      cout << endl << "/" << nb_s << ":"; 
  }

  /************   We find the permut' possible a priori (betas), and store them in Alpes  ****************/
  bool not_a_good_matrix = false;

  list< pair<int,int> > Imax = find_biggests(M);
  if ((Imax.size() != 2) && (verbose_multi)){ cout << TextMauve << "["; cout.flush(); }
  for (list< pair<int,int> >::iterator choice_bigger = Imax.begin(); choice_bigger != Imax.end(); ++choice_bigger){
  
    vector<double> combines_y = extract_y(M, (*choice_bigger).first);
    vector<double> betas_y2 = find_betas_graph(combines_y);
    not_a_good_matrix = (betas_y2.size() == 0);
    if (not_a_good_matrix)
      continue;
    //      return Alpes; // il est encore vide.
    //  if (int(betas_y2.size()) > 1){
    //cout "Matrix magique !" << endl;
    //M.display_text();
    //    //cout << "Betas_y2 : "; display_text(betas_y2);
    //  }
    
    pair< vector<int>, vector<vector<int> >   >  Permut_y;
    if (check_similar_lines){
      Permut_y = find_permutations_and_sort(betas_y2, combines_y, similar_columns);
      similar_columns.clear();
    }
    else
      Permut_y = find_permutation(betas_y2, combines_y);

    combines_y.clear();
    DoubleImage Mp = permute_y(M, Permut_y.first);



    vector<double> combines_x = extract_x(M, (*choice_bigger).second);
    vector<double> betas_x2 = find_betas_graph(combines_x);
    if (betas_x2.size() == 0)
      not_a_good_matrix = true;
    if (not_a_good_matrix)
      continue;
    //      return Alpes; // il est vide.

    //  if ( (int(betas_y2.size()) > 1) || (int(betas_x2.size()) > 1) ){
    //    cout << "Matrix magique !" << endl;
    //    M.display_text();
    //  }
    
    pair< vector<int>, vector<vector<int> >   >  Permut_x;
    if (check_similar_lines){
      Permut_x = find_permutations_and_sort(betas_x2, combines_x, similar_lines);
      similar_lines.clear();
    }
    else
      Permut_x = find_permutation(betas_x2, combines_x);

    combines_x.clear();
    Permut_x.first = compose_permutations(Permut_x.first, reverse(Permut_x.first.size()));
    reverse_index(Permut_x.second, Permut_x.first.size());
    Mp = permute_x(Mp, Permut_x.first);

    /************ rordonnancements possibles *********/
    DoubleImage alpha( betas_x2.size(), betas_y2.size() );
    betas_y2.clear();
    betas_x2.clear();


    for (int i = 0; i<alpha.nx; i++)
      for (int j = 0; j<alpha.ny; j++)
	alpha(i,j) = matrix_to_alpha(Mp, i, j);
  
    if (is_in_the_family(Mp,alpha)){
      Alpes.push_back( AlphaPermut(alpha, Permut_x.first, Permut_y.first, false ) );
    }


    if  (   ( Permut_x.second.size() != 0 ) ||  ( Permut_y.second.size() != 0 )    ){ // ambiguities : deal all permut'

      int np_tot = Permut_x.second.size() + Permut_y.second.size();
      int np_1 = Permut_x.second.size();
      vector<int>  coord(np_tot, 0);     // table-coordinates of a permutation
      vector<bool> length_cycles(max(Mp.nx, Mp.ny)+1, false); // we list the generic permutations to prepare
      
      long int nbp = 1;
      for (int k = 0; k<np_1; k++){
	length_cycles[ Permut_x.second[k].size() ] = true;
	nbp *= factorielle( Permut_x.second[k].size() );
      }
      for (int k = 0; k<int(Permut_y.second.size()); k++){
	length_cycles[ Permut_y.second[k].size() ] = true;
	nbp *= factorielle( Permut_y.second[k].size() );
      }
      //    if (nbp > 500){
      if (verbose_multi){ cout << TextMauve << "*" << nbp; cout.flush(); }
      //    } else {
      //    cout << "-" << nbp; cout.flush();
      //    }
      // if (nbp > 100){
      //   Mp.display_text();
      //   alpha.display_text();
      // }

      vector< vector< vector<int> > > Perm_pures = generate_all_permutations(length_cycles);
      vector< vector< vector<int> > > Suite_perm(Perm_pures.size());
    
      for (int k = 0; k<int(Perm_pures.size()); k++)
	if (length_cycles[k])
	  Suite_perm[k] = sequence_of_permutations(Perm_pures[k]);
      
      do {
	
	int deg = 0;
	bool cran_de_plus;
	do {
	  cran_de_plus = false;
	  
	  if (deg<np_1){
	    apply_permut_index_x(Mp, Permut_x.second[deg], Suite_perm[ Permut_x.second[deg].size() ][ coord[deg] ] );  // matrix, indices, permut
	    coord[deg]++;
	    if (coord[deg] ==  int( Suite_perm[ Permut_x.second[deg].size() ].size() )  ){
	      cran_de_plus = true;
	      coord[deg] = 0;
	      deg++;
	    }
	  } else {
	    apply_permut_index_y(Mp, Permut_y.second[deg-np_1], Suite_perm[ Permut_y.second[deg-np_1].size() ][ coord[deg] ] );  // matrix, indices, permut
	    coord[deg]++;
	    if (coord[deg] ==  int( Suite_perm[ Permut_y.second[deg-np_1].size() ].size() ) ){
	      cran_de_plus = true;
	      coord[deg] = 0;
	      deg++;
	    }
	  }
	  
	}
	while (cran_de_plus && (deg<np_tot));
	
	if (cran_de_plus) // all permut' have been tested, back to the initial state
	  break;
	
	
	for (int i = 0; i<alpha.nx; i++)
	  for (int j = 0; j<alpha.ny; j++)
	    alpha(i,j) = matrix_to_alpha(Mp, i, j);
	

	if (is_in_the_family(Mp,alpha)){
	  vector<int> coord_x ( coord.begin(), coord.begin()+np_1 );
	  vector<int> coord_y ( coord.begin() + np_1, coord.end() );
	  vector<int> px =  permutation_coordonnee(Mp.nx, coord_x, Permut_x.second, Perm_pures);
	  vector<int> py =  permutation_coordonnee(Mp.ny, coord_y, Permut_y.second, Perm_pures);
	  coord_x.clear(); coord_y.clear();
	  Alpes.push_back( AlphaPermut(alpha, 
				       compose_permutations(Permut_x.first, px), 
				       compose_permutations(Permut_y.first, py), 
				       false ) );
	}
	
	
	
      } while (true);
      
      Suite_perm.clear();
      Perm_pures.clear();
      
      //    if (nbp > 500){
      if (verbose_multi){ cout << "." << TextNormal; cout.flush(); }
      //    }
    }
    
    
    //cout << "Matrix " <<  Mp.nx << "x" <<  Mp.ny << ".  Ambiguities : " << Permut_x.second.size() << " x " << Permut_y.second.size() << ".  Nb possible : " << Alpes.size() << endl;
    
    
    alpha.empty();
    Mp.empty();
    
  } // End of loop on location of biggest element
  if ( (Imax.size() != 2) && (verbose_multi) ) { cout << TextMauve << "]" << TextNormal; cout.flush(); }
  
  return Alpes;
}




/******************************************* Canonization ****************************************/

// Canonises any matrix. Puts the sums over lignes in Vx and the sums over columns in Vy.
void canonise(DoubleImage& M, vector<double>& Vx, vector<double>& Vy){
  assert(M.nx == int(Vx.size()));
  assert(M.ny == int(Vy.size()));

  // pass over lines
  for (int i = 0; i<M.nx; i++){
    double a = 0;
    for (int j = 0; j<M.ny; j++)
      a += M(i,j);
    a /= M.ny;
    Vx[i] += a;
    for (int j = 0; j<M.ny; j++)
      M(i,j) -= a;
  }

  // passe over columns
  double b = 0;
  for (int j = 0; j<M.ny; j++){
    double a = 0;
    for (int i = 0; i<M.nx; i++)
      a += M(i,j);
    a /= M.nx;
    Vy[j] += a;
    b += a;
    for (int i = 0; i<M.nx; i++)
      M(i,j) -= a;
  }

  // global final pass
  b /= M.ny;
  for (int i = 0; i<M.nx; i++)
    for (int j = 0; j<M.ny; j++)
      M(i,j) -= b;

  for (int i = 0; i<M.nx; i++)
    Vx[i] += b;

}


// Removes the mean value of a vector (of potentials)
double canonise(vector<double>& V){
  double a = 0;
  int n = V.size();
  for (int i = 0; i<n; i++)
    a += V[i];
  a /= n;
  for (int i = 0; i<n; i++)
    V[i] -= a;
  return a;
}




/******************************* Random stuff ***************************/

// Generates a random matrix with coefficients between 0 and 1 (uniform law)
DoubleImage random_matrix(int n, int m, double coeff = 1.){
  DoubleImage M(n,m);
  for (int i = 0; i<n; i++)
    for (int j = 0; j<m; j++)
      M(i,j) = rand() * coeff / RAND_MAX;
      //      M(i,j) = (rand() % 2 )* 1. / 2;

  return M;
}


// Generates a random vector with coefficients between 0 and 1 (uniform law)
vector<double> random_vector(int n, double coefficient = 1.){
  vector<double> V(n);
  for (vector<double>::iterator i = V.begin(); i != V.end(); ++i)
    (*i) = rand() * coefficient / RAND_MAX;

  return V;
}



/********************** MRF representation ******************/

// Structure that encodes an MRF
struct Energy {
  int nb_var; // nb variables
  double constante; // additional constant
  vector< vector<double> > Pot; // the potentials of each label for each variable
  vector< vector< pair<int, int> > > Neighbors; // list, for each variable, of the indices of its neighbors (with the index of the associated interaction matrix), initialized to "no neighbor"
  typedef vector< pair<DoubleImage, pair<int,int> > >   MIV; // type for what follows
  MIV Interact; // interaction matrices, with the indices of the variables involved

  Energy(){
    nb_var = 0;
    constante = 0;
  }

  // Warning, this constructor copies pointers to the tables of interaction matrices
  Energy(const vector< vector<double> >& P, const vector< vector< pair<int, int> > >& V, const MIV& I, double c = 0) : constante(c), Pot(P), Neighbors(V), Interact(I) {
    assert(Pot.size() == Neighbors.size());
    nb_var = Pot.size();
  }

  // Copies interaction matrices (element by element, not just pointers)
  void copy(const Energy& A){
    nb_var = A.nb_var;
    constante = A.constante;
    Pot = A.Pot;
    Neighbors = A.Neighbors;
    Interact = vector< pair<DoubleImage, pair<int,int> > >(A.Interact.size());
    for (int i = 0; i< int(Interact.size()); i++){
      Interact[i].second = A.Interact[i].second;
      MRF_F0::copy(A.Interact[i].first, Interact[i].first);
    }
  }

  ~Energy(){;} // destructor by default : does not free the matrices, but frees the rest

  void empty(){
    for (int i = 0; i< int(Interact.size()); i++)
      Interact[i].first.empty();

    Interact.clear();
    Pot.clear();
    Neighbors.clear();
  }

  void display_text(int precis = 5) const{

    cout << "Cnste : " << constante << endl << endl;
    for (int i = 0; i<nb_var; i++){
      cout << "Pot " << i << " :  "; MRF_F0::display_text(Pot[i]);
    }
    
    for(MIV::const_iterator i = Interact.begin(); i != Interact.end(); ++i){
      cout << "Int. " << (*i).second.first << " x " << (*i).second.second << ":" << endl;
      (*i).first.display_text(precis);
    }

  }

};



// Evaluates the energy associated to a state (a choice of all labels)
double evaluate_energy(const Energy& A, const vector<int>& S){
  double v = A.constante;
  assert(S.size() == A.Pot.size());
  vector<int>::const_iterator j = S.begin();
  for (vector< vector<double> >::const_iterator i = A.Pot.begin(); i != A.Pot.end(); ++i){
    v += (*i)[*j];
    ++j;
  }

  for (Energy::MIV::const_iterator k = A.Interact.begin(); k != A.Interact.end(); ++k){
    v += (*k).first( S[ (*k).second.first ], S[ (*k).second.second ] );
  }

  return v;
}



/***************************** The family F0 and permutations ************************/

// Generates a matrix which is in the family F0 after permutation
DoubleImage generate_possible_matrix(int n, int m, bool permut = true, bool sort_sums = false){
  DoubleImage alphas_init = generate_alphas(n,m);  // positive !
  if (sort_sums)
    alphas_init = sort_decreasing_sums(alphas_init);
  DoubleImage Mat_init = alphas_to_matrix(alphas_init);
  alphas_init.empty();

  if (permut){
    vector<int> permut_init_x = generate_permutation(Mat_init.nx);
    vector<int> permut_init_y = generate_permutation(Mat_init.ny);
    DoubleImage M = permute_x(Mat_init, permut_init_x);
    Mat_init.empty();
    permut_init_x.clear();
    M = permute_y(M, permut_init_y);
    permut_init_y.clear();
    return M;
  } else {
    return Mat_init;
  }
}



// Check whether a permutation of the binary-formatted labels is a permutation of the nodes,
// and if yes, returns the node permutation (last bit in a binary-formatted label = 0-th node, first bit = last node);
// otherwise, returns a vector of length 0.
vector<int> find_permutation_nodes(const vector<int>& P){
  int nn = int( log(P.size()) / log(2) + 0.1); // nb nodes
  assert( int( pow(2, nn) + 0.1 ) == int( P.size() ) );
  vector<int> A(nn); // permut' of nodes, since we already have the permut' of labels...


  // Looks at the image of each node, by checking the 00010000 (i.e. all 0*10*)
  // We check also that it is one-to-one (= bijective).
  vector<bool> bij(nn,false);
  for (int b = 0; b<nn; b++){
    int s = int( pow(2, b) + 0.1 );
    int t = P[s];
    
    if (t == 0){
      A.clear();
      bij.clear();
      return A;
    }
      
    int nt = int( log(t) / log(2) + 0.1);
    if ( (int( pow(2,nt) + 0.1 ) != t)  ||  ( bij[nt] ) ){ // pb if the image of the node is not a node, or if the image has already been produced before
      bij.clear();
      A.clear();
      return A;
    }
    
    A[b] = nt;
    bij[nt] = true;
  }
  bij.clear();

  // Generates the permutation over all labels and checks it works
  for (int i = 0; i<int(P.size()); i++){
    int k = 0;
    for (int j = 0; j<nn; j++){
      int jbit = (i / ( int(pow(2,j) + 0.1) ) ) % 2; // j-th bit
      if (jbit == 1)
	k += int( pow(2, A[j]) + 0.1);
    }
    
    if (k != P[i]){ // does not generate the right permut'...
      A.clear();
      return A;
    }
  }
  
  return A;
}




/********************* Switching between alphas and vectors/matrices *****************/


// Given a variable x_i and a way to interpret it (= the permutation : binary format --> {1, 2, 3...} order of the labels),
// returns the decomposition of the potential of this variable (the alpha_i and alpha_ij of the definition).
// If "empty" is true, then returns a vector of length 0 if the coefficients of the coupled part are not positive, or if the reconstruction failed. 
// Otherwise, returns the coefficients, whatever they are, without checking reconstruction (i.e. that the original potential might be not representable and thus differ from the reconstructed ones, but this is not checked in this case).
vector<double> decompose_potential(const vector<double>& Pot,  const vector<int>& Interp, bool empty = false){

  assert( Pot.size() == Interp.size() );
  int n_labl = Pot.size();
  int na_pur = int( log(n_labl) / log(2) + 0.1);
  assert(int( pow(2,na_pur) + 0.1) == n_labl);

  int na_tot = (na_pur * (na_pur + 1 ) ) / 2 ; // n pure + n(n-1)/2 coupled
                                               //  in R,     in R+

  vector<double> Pot2 = permute(Pot, Interp);

  vector<double> Coeffs(na_tot,0);

  for (int c = 0; c < na_pur; c++){ // the coeffs in R
    for (int i = 0; i < n_labl; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      Coeffs[c] += s * Pot2[i];
    }
    Coeffs[c] /= n_labl;
  }


  int cn = na_pur;
  for (int c1 = 0; c1 < na_pur; c1++) // the coeffs in R+
    for (int c2 = c1+1; c2 < na_pur; c2++){

      for (int i = 0; i < n_labl; i++){
	int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
	Coeffs[cn] += s * Pot2[i];
      }

      if (empty){ // we test whether coeff > 0.
	if (Coeffs[cn] < 0){
	  Pot2.clear();
	  Coeffs.clear();
	  return Coeffs;
	}
      }
      
      Coeffs[cn] /= n_labl;
      cn++;
    }
  assert(cn == na_tot);



  if (empty){ // we test reconstruction
    vector<double> A(n_labl, 0);

    for (int c = 0; c < na_pur; c++)
      for (int i = 0; i < n_labl; i++){
	int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
	A[i] += Coeffs[c] * s;
      }
    
    int cr = na_pur;
    for (int c1 = 0; c1 < na_pur; c1++) 
      for (int c2 = c1+1; c2 < na_pur; c2++){
	for (int i = 0; i < n_labl; i++){
	  int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
	  A[i] += Coeffs[cr] * s;
	}
	cr++;
      }

    if (! is_equal(A, Pot2)){
      Pot2.clear();
      Coeffs.clear();
      A.clear();
      return Coeffs;
    }

    A.clear();
  }
  Pot2.clear();

  return Coeffs;
}



// Generation of a possible vector of potentials
// No additional permutations here.
vector<double> generate_possible_potential(int n_labl){

  int na_pur = int( log(n_labl) / log(2) + 0.1); // !!! if log(8)/log(2) = 2.99999 then int(2.999...) returns 2 !
  assert(int( pow(2,na_pur) + 0.1) == n_labl);

  //  int na_tot = (na_pur * (na_pur + 1 ) ) / 2 ; // n purs + n(n-1)/2 couples
  //                                               //  in R,    in R+

  vector<double> A(n_labl, 0);

  for (int c = 0; c < na_pur; c++){
    double coeff = 2 * ( (rand() * 1. /  RAND_MAX) - 0.5) ; // between -1 et 1
    for (int i = 0; i < n_labl; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      A[i] += coeff * s;
    }
  }    

  for (int c1 = 0; c1 < na_pur; c1++) 
    for (int c2 = c1+1; c2 < na_pur; c2++){
      double coeff = rand() * 1. / RAND_MAX ; // between 0 et 1
      for (int i = 0; i < n_labl; i++){
	int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
	A[i] += coeff * s;
      }
    }
  
  return A;
}





// Re-creates the vector of potentials from the alphas and their interpretation
// Returns V_alpha * Interp^-1
vector<double> alphas_and_interp_to_potential(const vector<double>& Alpha, const vector<int>& Interp){
  
  int n_labl = int(Interp.size()); // nb labels
  int n = int( log(n_labl) / log(2) + 0.1); // nb nodes
  assert( int( pow(2,n) + 0.1) == n_labl );
  assert( (n*(n+1))/2 == int(Alpha.size()) );

  //  int n_tot = (n * (n + 1 ) ) / 2 ; // n pure + n(n-1)/2 couples
  //                                           // in R,  in R+

  vector<double> V(n_labl, 0);

  for (int c = 0; c < n; c++)
    for (int i = 0; i < n_labl; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      V[ Interp[i] ] += Alpha[c] * s;
    }
  
  int cr = n;
  for (int c1 = 0; c1 < n; c1++) 
    for (int c2 = c1+1; c2 < n; c2++){
      for (int i = 0; i < n_labl; i++){
	int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
	V[ Interp[i] ] += Alpha[cr] * s;
      }
      cr++;
    }
  
  return V;
}



// Recreates the vector of potentials from its alphas (interp = id)
// Returns V_alpha
vector<double> alphas_to_potential(const vector<double>& Alpha){
  
  int n = int( sqrt(2. * Alpha.size()) );
  int n_labl = int( pow(2,n) + 0.1);
  assert( (n*(n+1))/2 == int(Alpha.size()) );

  //  int n_tot = (n * (n + 1 ) ) / 2 ; // n pure + n(n-1)/2 couples
  //                                           // in R,  n R+

  vector<double> V(n_labl, 0);

  for (int c = 0; c < n; c++)
    for (int i = 0; i < n_labl; i++){
      int s = ( ith_bit(c,i) == 0 )?(-1):(1) ;
      V[ i ] += Alpha[c] * s;
    }
  
  int cr = n;
  for (int c1 = 0; c1 < n; c1++) 
    for (int c2 = c1+1; c2 < n; c2++){
      for (int i = 0; i < n_labl; i++){
	int s = ( ith_bit(c1,i) + ith_bit(c2,i) == 1 )?(1):(-1) ;
	V[ i ] += Alpha[cr] * s;
      }
      cr++;
    }
  
  return V;
}




/*************************** Coherency of representations of interacting variables *****************/

// Structure to explore global coherencies of the representation of potentials and interactions
struct Exploration {
    
  list<int> *  To_be_explored; // nodes to be explored, i.e. neighbors of nodes already explored
  list< pair<int,int> > * To_be_explored_later; // neighbors to be explored, but with several possible interpretations (so we postpone their study, hoping that the study of other common neighbors will select just one possible interpretation [or just a few]).
  vector<bool> * Already_explored; // true if already explored, false otherwise : we check this table before adding a node in a list or before analyzing it.
  int nb_already_explored; // nb of 'true' in the table above.
  vector<int> * Interp_chosen; // Interp_chosen[i] is the index N of the interpretation chosen for variable x_i in Coherence[i][N]
  // Coherence[i][ Interp_chosen[i] ].first : interpretation of x_i
  
  Exploration(){
    To_be_explored = NULL; To_be_explored_later = NULL; Already_explored = NULL; Interp_chosen = NULL; 
    nb_already_explored = 0; 
  }

  Exploration(list<int>* A, list< pair<int,int> >* B, vector<bool>* C, int f, vector<int>* D) : To_be_explored(A), To_be_explored_later(B), Already_explored(C), nb_already_explored(f), Interp_chosen(D) {;}

  Exploration(list<int>& A, list< pair<int,int> >& B, vector<bool>& C, int f, vector<int>& D) : To_be_explored(&A), To_be_explored_later(&B), Already_explored(&C), nb_already_explored(f), Interp_chosen(&D) {;}

  Exploration(const list<int>& A, const list< pair<int,int> >& B, const vector<bool>& C, int f, const vector<int>& D, bool b){
    b = true;

    To_be_explored = new list<int>;
    *To_be_explored = A;
    To_be_explored_later = new list< pair<int,int> >;
    *To_be_explored_later = B;
    Already_explored = new vector<bool>;
    *Already_explored = C;
    nb_already_explored = f;
    Interp_chosen = new vector<int>; 
    *Interp_chosen = D;
  }

  //destructor only empty pointers, while here we empty data too.
  void empty(){
    if (To_be_explored != NULL)
      To_be_explored->clear();
    
    if (To_be_explored_later != NULL)
      To_be_explored_later->clear();

    if (Already_explored != NULL)
      Already_explored->clear();

    if (Interp_chosen != NULL)
      Interp_chosen->clear();
    
    To_be_explored = NULL; To_be_explored_later = NULL; Already_explored = NULL; Interp_chosen = NULL;
    nb_already_explored = 0;
  }

};
  



// Tests experimentally whether 2 energies are equal
// by estimating the 2 energies at nb_tests random states of the variables
bool test_random_equality_energies(const Energy& E0, const Energy& E1, int nb_tests = 100, double epsi = 0.00000001){
  
  for (int k = 0; k<nb_tests; k++){
    vector<int> etat( E0.nb_var ); // a state of the variables
    for (int i = 0; i<int(etat.size()); i++)
      etat[i] = rand() % int( E0.Pot[i].size() );
    
    double v0 = evaluate_energy(E0, etat);
    double v1 = evaluate_energy(E1, etat);
    if ( fabs(v0 - v1) > epsi ){
      cout << "Values energies differentes !  " << v0 << " vs " << v1 << "  pour l'etat "; display_text(etat);
      return false;
    }
  }
  
  return true;
}




/******************** Energy representation *******************/

// Canonises an energy
void canonise(Energy& E){
  
  for (Energy::MIV::iterator i = E.Interact.begin(); i != E.Interact.end(); ++i)
    canonise( (*i).first, E.Pot[ (*i).second.first ], E.Pot[ (*i).second.second ] );
  
  //    cout << "Energy E1 : " << endl;
  //    E.display_text();
  
  for (vector< vector<double> >::iterator i = E.Pot.begin(); i != E.Pot.end(); ++i)
    E.constante += canonise(*i);
  
}



// Structure to store the information required for the construction of the graph, in the case of a unique interpretation
struct PreGraph {
  vector< DoubleImage > Alpha; // contains the representation alpha of interaction matrices
  vector< vector<int> > Interp; // Local (unique) intepretation 
  vector< vector<double> > Coeff_Pot; // Decompositions of potentials according to these interpretations
  vector< bool > solitary; // 'true' for variables not depending on any other one (i.e. forming a connected component of the graph).
  
  ~PreGraph(){ // to empty what does not empty by itself.
    empty();
  }
  
  void empty(){
    for (int i = 0; i< int(Alpha.size()); i++)
      Alpha[i].empty();
  }
  
};




/******************* Graph structures ********************/


// Structure to store the information required for the construction of the graph, in the case of multiple local interpretations found and of a choice of these interpretations
struct PreGraphMulti {
  vector<  vector< AlphaPermut > > AlpesL; // contains the possible representations (alpha, px, py) of all interaction matrices
  vector< vector<Interpretation> > Coherence; // Possible local intepretations 
  vector<  vector< vector<double> > > Coeff_Pot; // Decompositions of potentials according to these interpretations
  vector< bool > solitary; // 'true' for variables not depending on any other one (i.e. forming a connected component of the graph)
  list< vector<int> > Solutions; // global solutions : each solution is just a choice of interpretation (= an int) for each x_i
  
  ~PreGraphMulti(){ // to empty what does not empty by itself.
    empty();
  }
  
  void empty(){
    for (int i = 0; i< int(AlpesL.size()); i++)
      for (int j = 0; j< int(AlpesL[i].size()); j++)
	AlpesL[i][j].empty();
  }
  
};





// Builds the pre-graph...
 void construct_pregraph(PreGraphMulti& PG, const Energy& E, bool verbose_base = false, bool verbose_a_bit_more = false, bool verbose_multi = false, bool verbose_coherence = false, bool verbose_coherence_pot = false, bool verbose_global = false, ostream& cout = std::cout){
  
  if (verbose_base)
    cout << "Find alpha and permutations back" << endl;
  
  if (verbose_a_bit_more)
    cout << "Summary of the search (nb of local possibilities for interaction matrices) :" << endl;
  
  
  
  // The decompositions of the matrices as (alpha, permut')   // AlpesL = List of Alpha-Permut
  vector<  vector< AlphaPermut > >& AlpesL = PG.AlpesL;
  AlpesL.resize( E.Interact.size() ); // store the permutations and alphas corresponding to the interaction matrices
  vector<  vector< AlphaPermut > >::iterator ialp = AlpesL.begin();
  for (Energy::MIV::const_iterator imiv = E.Interact.begin(); imiv != E.Interact.end(); ++imiv){
    *ialp = find_alpha_and_permutation( (*imiv).first, false, verbose_multi);
    //      *ialp = find_alpha_and_permutation( (*imiv).first, true);
    
    
    switch (int( (*ialp).size() ) ) {
    case 0: 
      //	cout << "Matrix not possible !!!!!" << endl;
      if (verbose_a_bit_more)
	cout << "("; cout.flush();
      break;
    case 1 :
      if (verbose_a_bit_more)
	cout << "^"; cout.flush();
      break;
    default:
      if (verbose_a_bit_more){ cout << (*ialp).size() << "_"; cout.flush(); }
    }
    
    
    ++ialp;
  }


  /*************************************************************************************************************/
  /************** we search whether there exist coherent permutations between interaction matrices *************/
  /*************************************************************************************************************/
  
  if (verbose_base)
    cout << endl << "Local coherency" << endl;
  
  if (verbose_a_bit_more)
    cout << "Summary (nb of locally-possible representations compatible with neighbors) :" << endl;
  
  
  // Solitary variables
  PG.solitary = vector<bool>(E.nb_var, false);

  // interpretation of a variable x_i
  //      interpretation de x_i,  interpretation of lines of interaction matrices with neighbors (index of the label permutation, and node permutation).
  //    typedef     pair< vector<int>,  vector< pair<int, vector<int> > > > Interpretation;
  
  vector< vector<Interpretation> >&  Coherence = PG.Coherence;
  Coherence.resize(E.nb_var); // for each variable, list of locally-coherent interpretations (i.e. possible interaction matrices)
  
  // coefficients of the decompositions of the potentials E.Pot[i] of each variable x_i, for each interpretation locally coherente of x_i
  vector<  vector< vector<double> > >& Coeff_Pot = PG.Coeff_Pot;
  Coeff_Pot.resize(E.nb_var);

  
  for (int i = 0; i<E.nb_var; i++){ // loop over variables
    
    // Future interpretations of x_i, and their decompositions
    // For each possible interpretation, a list of n + n(n-1)/2 = n(n+1)/2 coefficients (where 2^n = #X_i)
    vector<  vector<double>  >& Alain = Coeff_Pot[i]; // Alpha-Interpr
    
    if (verbose_coherence)
      cout << "- Variable x_" << i << endl;

    // Recall : interpretation of a variable x_i
    //	typedef     pair< vector<int>,  vector< pair<int, vector<int> > > > Interpretation;
    //      interpretation of x_i,  interpretation of lines of interaction matrices with neighbors (index of label permut', and node permut').
    vector< Interpretation >& Inter_pos = Coherence[i];
    
    int nbv = int(E.Neighbors[i].size()); // nb of neighbors of x_i
    if (nbv == 0){ // situation a bit silly but that still should be dealt with
      
      //	if (verbose_coherence_pot)
      if (verbose_a_bit_more)
	cout << "!!!!!!!!!!!!!!!!!!!  One may remove variable x_" << i << " from the minimization problem since it is independent from the other ones !!!!!!!!!!!!!!!!!!!" << endl;
      
      // Then any order is fine, not important for the graph cut
      
      Inter_pos.push_back( make_pair( identity_permutation(int(E.Pot[i].size())),   vector< pair<int, vector<int> > >() ) );
      PG.solitary[i] = true;
      
    } else { // sufficient to check compatibility of each interpretation due to an interaction matrix with another of reference
      
      
      // Better order by increasing number of possible permutations, in order to avoid useless computations.
      // Or at least consider as reference an interaction matrix with one only possible interpretation
      int iv_ref = 0;
      int nperv2 = AlpesL[ E.Neighbors[i][0].second ].size();
      
      for (int j = 1; j<nbv; j++){
	int nperv =  AlpesL[ E.Neighbors[i][j].second ].size();
	
	if (nperv < nperv2){
	  nperv2 = nperv;
	  iv_ref = j;
	}
	
	if (nperv == 1)
	  break;
      }
      
      //	int v_ref =  E.Neighbors[i][iv_ref].first; // the iv_ref-th neighbor is node number v_ref
      int num_inter_ref = E.Neighbors[i][iv_ref].second; // interaction matrix i <--> v_ref, beware the reading order 
      if (verbose_coherence)
	cout << "  \\- Neighbor ref: number " << iv_ref << ", matrix num : " << num_inter_ref << endl;
      
      
      vector < AlphaPermut >& Perm_ref = AlpesL[num_inter_ref];
      
      
      // For a given variable x_i :
      //  set of possible perm',      the permut' = interpretation         
      //	vector<                      pair<  vector<int> ,
      // neighbors   permut q of neighbor : only one possible,   permut' of nodes
      //  vector<       pair<  int ,                               vector<int>             >    >  > >  Inter_pos;
      // unique up to node permutation for interpretation.
      
      // Inter_pos : list of possible interpretations for variable x_i
      // Inter_pos[k] : k-ith possible interpr
      // Inter_pos[k].first : permutation-interpretation number k
      // Inter_pos[k].second: node permutations to apply to interaction matrices with neighbors of x_i
      // Inter_pos[k].second.size() : nb of neighbors of x_i
      // Inter_pos[k].second[j] : permut' corresponding to neighbor j
      // Inter_pos[k].second[j].first : index of the right AlphaPerm  among the ones possible for interaction matrix with neighbor j
      // Inter_pos[]k].second[j].second: node permutation to apply to (the right) AlphaPerm with neighbor j (to its px or to its py)
      
      // interact. mat.  btw. x_i & its j-th neighb.         selection of the right AlphaPerm
      //        AlpesL[     E.Neighbors[i][j].second     ][    Inter_pos[i].second[j].first   ] -> px/py  according to (i < E.Neighbors[i][j].first)
      
      
      for (int k = 0; k < int(Perm_ref.size()); k++){ // loop over possible interpretations of the interaction matrix (x_i, x_neighb_ref) with the fewest ones
	
	// For a given interpretation interp_k of M_{node,v_ref},
	// perm_kept(num_voisin) = list of compatible pairs (index of interpretation of M_{node, neighb}, the node permut' going with)
	vector<   list < pair<  int, vector<int>  > >  >  perm_kept( E.Neighbors[i].size() );
	if (verbose_coherence)
	  cout << "  \\- We test a new initial permutation of x_i=" << i << ", over " << Perm_ref.size() << endl;
	
	for (int j = 0; j<int(E.Neighbors[i].size()); j++){ // loop over neighbors (except reference one)
	  
	  if (j == iv_ref){ 
	    perm_kept[iv_ref].push_back( make_pair(k, identity_permutation( int( log(E.Pot[i].size()) / log(2) + 0.1 ) ) ) );
	    continue;
	  }
	  
	  
	  if (verbose_coherence)
	    cout << "    \\- Neighbor index j = " << j << " / " << nbv << endl;
	  
	  for (int kj = 0; kj< int( AlpesL[ E.Neighbors[i][j].second ].size() ); kj++){    // loop over possible interpretations of matrix (x_i, x_neighb_num_j)
	    
	    if (verbose_coherence)
	      cout << "      \\- Test of permutation of x_j num " << kj << " / " << int( AlpesL[ E.Neighbors[i][j].second ].size() ) << endl;
	    
	    vector<int> composee = compose_permutations(    inverse_permutation( *(  AlpesL[ E.Neighbors[i][j].second ][ kj ].p(E.Neighbors, i, j) ) ), 
							    *(   Perm_ref[k].p(E.Neighbors, i, iv_ref)  )       );
	    
	    if (verbose_coherence){ cout << "       - Vector composed : " << composee << endl; }
	    // M_alpha_{i,vj} o composee o M_alpha_{i,ref}.px^-1 = M_orig{i,vj} {1,2,3...}
	    vector<int> perm_node = find_permutation_nodes(composee);
	    if (verbose_coherence){ 
	      if (perm_node.size() == 0)
		cout << "       - not a node perm" << endl;
	      else { cout << "       - Node perm : " << perm_node << endl;; } 
	    }
	    
	    if ( perm_node.size() != 0 ){ // compatible 
	      perm_kept[j].push_back( make_pair( kj, perm_node ) );
	    }
	  }
	  
	  if ( perm_kept[j].size() == 0 ){ // ie. if no comptabile interpretation found for the j-th neighbor
	    perm_kept.clear(); // then we can empty this interpretation for all neighbors
	    break;
	  }
	} // end loop over neighbors
	
	
	/*********** We check if the V_i can we written as they are via the interpretations *********/
	
	
	if (perm_kept.size() != 0){
	  if (verbose_coherence_pot){
	    cout << endl << "   \\- Coherency of Pot[i=" << i << "] : "; cout.flush();
	  }	    
	  vector<double> coeff = decompose_potential(E.Pot[i],  *( Perm_ref[k].p(E.Neighbors, i, iv_ref) ) , true );
	  //	    cout << "Poten: "; display_text(E.Pot[i]);
	  //	    cout << "Inter: "; display_text( *( Perm_ref[k].p(E.Neighbors, i, iv_ref) ) );
	  if (coeff.size() != 0){
	    if (verbose_coherence_pot)
	      cout << "ok" << endl;
	    //	      cout << "Coeff: "; display_text(coeff);
	    
	    int nbv = int(perm_kept.size()); assert(nbv == int(E.Neighbors[i].size())); //nb neighbors
	    vector< list< pair< int, vector<int> > >::iterator > coord(nbv);
	    for (int j = 0; j<nbv; j++)
	      coord[j] = perm_kept[j].begin(); 
	    
	    do { // loop over possible combinations of interpretations
	      
	      vector< pair< int, vector<int> > > perm_to_add(nbv);
	      for (int j = 0; j<nbv; j++)
		perm_to_add[j] = * coord[j];
	      
	      Inter_pos.push_back( make_pair( *( Perm_ref[k].p(E.Neighbors, i, iv_ref) ) , perm_to_add) );
	      Alain.push_back(coeff);
	      
	      int c = 0;
	      int tourner = false;
	      
	      do { // take the next combination
		++ coord[c];
		if (coord[c] == perm_kept[c].end()){
		  coord[c] = perm_kept[c].begin();
		  c++;
		  tourner = true;
		} else {
		  tourner = false;
		}
	      } while ((tourner) && (c < nbv));
	      if (tourner) // we looked all possible combinations of interpretations
		break;
	      
	    } while (true);
	    
	  } else if (verbose_coherence_pot)
	    cout << "not good" << endl;
	}
	
      } // End of test for the k-th interpretation of variable i
      
    } // end of test of coherence for variable x_i
    
    if (Inter_pos.size() == 0){
      //	cout << "Impossible energy..." << endl;
      Coherence.clear();
      break;
    }
    else
      if (verbose_coherence_pot)
	cout << Inter_pos.size() << " representations locally possible for variable x_" << i << endl;
      else {
	if (verbose_a_bit_more){  cout << "| x_" << i << " : " << Inter_pos.size() << " "; cout.flush(); }
      }
    
  } // end of loop over i
  
  if (Coherence.size() == 0){
    if (verbose_base)
      cout << endl << "Impossible energy..." << endl;
    else { 
      std::cout << BoldRed << "-X" << TextNormal; std::cout.flush(); 
    }
    
    return;
  }
  
  
  
  /********************************************************************************************/
  /******************* we search for globally coherent interpretations    *********************/
  /********************************************************************************************/
  
  
  if (verbose_base)
    cout << endl << "Energy locally possible, now we test global coherency..." << endl;
  
  //    bool verbose_global = true;
  
  //      interpretation of x_i,  interpretation of rows of interaction matrices with neighbors (index of the label permut', and node permut').
  //    typedef     pair< vector<int>,  vector< pair<int, vector<int> > > > Interpretation;

  //    vector< vector<Interpretation> > Coherence(E.nb_var); // for each variable, list of interpretations locally coherent (i.e. possible interaction matrices)
  
  // Each variable has an even number of locally possible interpretations, symmetric (by node inversion), one may explore just one interpretation of each pair for the first variable.
  // Choose a variable with few possible interpretations, just to start with... (to reduce complexity)
  int nb_inter_min = Coherence[0].size();
  int index_inter_min = 0;
  for (int i = 1; i<E.nb_var; i++){
    int nb2 = Coherence[i].size();
    if (nb2 < nb_inter_min){
      index_inter_min = i;
      nb_inter_min = nb2;
    }
    if (nb_inter_min == 2) // cannot do better
      break;
  }
  
  if (verbose_global)
    cout << endl << "Index first variable : x_" << index_inter_min << ", with " << nb_inter_min << " interpretations." << endl;
  
  list<Exploration> Exploits; // Successive explorations
  // Only one initial exploration
  
  list< vector<int> >& Solutions = PG.Solutions; // global solutions : each solution is just a choice of one interpretation (= an int) for each x_i
  
  
  list<int> _To_be_explored; // nodes to explore, i.e. neighbors of nodes already explored
  list< pair<int,int> > _To_be_explored_later; // neighbors to explore but with several possible interpretations (so we postpone, hoping the ambiguity will be resolved through other neighbors)
  vector<bool> _Already_explored(E.nb_var,false); // true if already explored, false otherwise : we check this table before adding a node in a list or before its analysis.
  vector<int> _Interp_chosen(E.nb_var, -1); // Interp_chosen[i] is the index N of the interpretation chosen for variable x_i in Coherence[i][N]
  // Coherence[i][ Interp_chosen[i] ].first : interpretation of x_i
  //    vector< AlphaPermut > _Alpins(AlpesL.size()); // list of node permutations (not labels!) to apply to matrix M_ij, with same indexation as AlpesL (px, py, but no alpha).
  
  _Already_explored[index_inter_min] = true;
  for (vector< pair<int,int> >::const_iterator i = E.Neighbors[index_inter_min].begin(); i != E.Neighbors[index_inter_min].end(); ++i){
    _To_be_explored.push_back( (*i).first );
    // Not obvious at first but it was not Identity
    //      _Alpins[ (*i).second ].p( ( index_inter_min < (*i).first) ) = new vector<int>( permutation_identite( int ( log(E.Pot[index_inter_min].size()) / log(2) + 0.1) ));
    
  }
  _Interp_chosen[index_inter_min] = 0;
  
  Exploration Dora(_To_be_explored, _To_be_explored_later, _Already_explored, 1, _Interp_chosen /*, _Alpins*/); // copy of pointers
  Exploits.push_back(Dora);
  
  for (int a = 1; a<nb_inter_min; a++){ // we have to add other possible explorations, just in case...
    Exploits.push_back( Exploration(_To_be_explored, _To_be_explored_later, _Already_explored, 1, _Interp_chosen,  true) ); // full re-copy of tables
    (* Exploits.back().Interp_chosen)[index_inter_min] = a;
  }
  
  while (! Exploits.empty()) { // As long as there remain explorations to be done...
    
    Exploration& Dorade = Exploits.front();
    int& nb_already_explored = Dorade.nb_already_explored;
    list<int>& To_be_explored = * Dorade.To_be_explored;
    list< pair<int,int> >& To_be_explored_later = * Dorade.To_be_explored_later;
    vector<bool>& Already_explored = * Dorade.Already_explored;
    vector<int>& Interp_chosen = * Dorade.Interp_chosen;
    //      vector<AlphaPermut>& Alpins = * Dorade.Alpins;
    
    if (verbose_global)
      cout << endl << endl << "*********************   New exploration !  ***************************" << endl << endl;
    
    do { // Loop on connected components...
      
      while ( (! To_be_explored.empty()) || (! To_be_explored_later.empty()) ) { // As long as there is something to do, let's go !
	
	
	if (verbose_global){
	  cout << "Nb already explored : " << nb_already_explored << endl;
	  cout << "To_be_explored : " << To_be_explored << endl;
	  cout << "To_be_explored_later : " << To_be_explored_later << endl;
	  cout << "Already explored : " << Already_explored << endl;
	  cout << "Interp_chosen : " << Interp_chosen << endl;
	}
	
	/***** find a "node" (= a variable x_i) with something to do *****/
	int node; 
	int mode_bourrin = false; // true if one has to remember the different possibilities instead of avoiding them when they are several.

	do {
	  
	  if (! To_be_explored.empty()){
	    node = To_be_explored.front();
	    To_be_explored.pop_front();
	  } else 
	    if (! To_be_explored_later.empty()){
	      if (!mode_bourrin) // It's our first time here
		To_be_explored_later.sort(sort_pair_2_croissant<int,int>); // increasing order of the 2nd element (the number of possibilities)
	      
	      node = To_be_explored_later.front().first;
	      To_be_explored_later.pop_front();
	      mode_bourrin = true;
	    } else {
	      node = -1;
	      break;
	    }
	  
	  if (! Already_explored[node]) // let's explore this one
	    break;
	  
	} while (true) ;
	  
	if (node == -1){
	  if (verbose_global)
	    cout << "No variable still unknown." << endl;
	  break;
	}	
	
	if (verbose_global){
	  cout << "- New variable to study : x_"<< node << endl;
	  cout << "- Berserk mode : "; 
	  if (mode_bourrin)
	    cout << "on";
	  else
	    cout << "off";
	  cout << endl;
	  
	  cout << "x_" << node << " a " << E.Neighbors[node].size() << " neighbors : " << E.Neighbors[node] << endl;
	}
	
	
	
	/******** We studay possibilities available for that node *********/
	// we keep the first one to remember it only in the end.
	int v_interp_k;
	//	vector< vector<int> > v_Perm_en_Stock;
	int v_trouve = 0;
	
	for (int interp_k = 0; interp_k < int( Coherence[node].size() ); interp_k++){ // loop over locally possible interpretations of "node"
	  
	  if (verbose_global)
	    cout << "--- We start studying interpretation " << interp_k << " / " << Coherence[node].size() << endl;
	  
	  //	  // we go though all neighbors of "node" and, for already explored ones, we check compatibility.
	  //	  // we store associated permutations (of nodes), when they exist, in a vector meant for.
	  //	  vector< vector<int> > Perm_en_Stock(E.Neighbors[node].size());
	  
	  bool compat = true; // set to false if ever a M_ij and a M_ji do not share the same representation
	  
	  for (int j = 0; j < int( E.Neighbors[node].size() ) ; j++){
	    
	    
	    int vj = E.Neighbors[node][j].first; // j-th neighbor of "node"
	    if (verbose_global)
	      cout << "   \\-- Neighbor index " << j << " / " << E.Neighbors[node].size() << " : " << vj << endl;
	    if ( ! Already_explored[vj] ){
	      if (verbose_global)
		cout << "     \\_ not studied yet (next one!)" << endl;
	      continue;
	    }	    
	    if (verbose_global)
	      cout << "     \\_ already studied" << endl;
	    
	    
	    //	    vector<int>& Interp_node = Coherence[node][ interp_k ].first ; // interp of index k of "node"
	    //	    vector< pair<int, vector<int> > >& Interp_neighbors_vj = Coherence[vj][ Interp_chosen[vj] ].second ; 
	    //	    //	    int node_pour_vj = find_value_1(Interp_neighbors_vj, node);
	    int node_pour_vj = find_value_1(E.Neighbors[vj], node);
	      
	    
	    
	    //	    // E.Neighbors[node][j].second ; // index of matrix M_{node,vj}
	    //	    //	Interp_neighbors_vj[node_pour_vj].first ; // index of representation of M_{node,vj} chosen by vj
	    //	    vector<int>& Interp_node_par_vj = * (AlpesL[ E.Neighbors[node][j].second ][ Interp_neighbors_vj[node_pour_vj].first ].p( (node<vj) ) );
	    //	    if (verbose_global){
	    //	      cout << "       \\_ Interp_node : " << Interp_node << endl;
	    //	      cout << "       \\_ Interp_nd_vj : " << Interp_node_par_vj << endl;
	    //	    }
	    //	    vector<int> composee = compose_permutations(    inverse_permutation( Interp_node_par_vj ), Interp_node ); // check sense
	    //	    Perm_en_Stock[j] = find_permutation_nodes(composee);
	    //	    if (Perm_en_Stock[j].size() == 0){
	    //	      // interpretations not compatible
	    //	      if (verbose_global)
	    //		cout << "       \\_ failure with neighbor (not compatible interpretations) : composed = " << composee << endl;
	    //
	    //	      Perm_en_Stock.clear();
	    //	      break;
	    //	    }
	    
	    
	    // we ask that the choice of representations of matrices M_ij should be the same as for interpretation of x_i as well as of x_vj...
	    if (  Coherence[node][ interp_k ].second[j].first   ==    Coherence[vj][ Interp_chosen[vj] ].second[node_pour_vj].first  ){
	      if (verbose_global)
		cout << "       \\_ compatible neighbor" << endl;
	    } else {
	      compat = false;
	      if (verbose_global)
		cout << "       \\_ neighbor failed : not compatible (" <<  Coherence[node][ interp_k ].second[j].first  << " vs " <<  Coherence[vj][ Interp_chosen[vj] ].second[node_pour_vj].first << ")" << endl;
	      break;
	    }
	    
	    
	  } // End loop neighbors
	  
	    //	  if (Perm_en_Stock.size() == 0){ // we try next interpretation of "node"
	  if (! compat){  // This interprtation of x_i does not work, let's try another one
	    if (verbose_global)
	      cout << "       \\_ next interpretation !" << endl;
	    continue;
	  }
	  
	  
	  // otherwise, either we are in berserk mode and store all possibilities,
	  //            either we're not, and we just count (and remember the first one if there's only one)
	  
	  if (v_trouve == 0){
	    v_trouve = 1;
	    v_interp_k = interp_k;
	    //	    v_Perm_en_Stock = Perm_en_Stock;
	    
	    if (verbose_global)
	      cout << "    --> first interpretation plausible" << endl;
	    
	  } else {
	    
	    if (mode_bourrin){
	      Exploits.push_back( Exploration(To_be_explored, To_be_explored_later, Already_explored, nb_already_explored, Interp_chosen /*, Alpins*/, true) ); // full copy of tables
	      Exploits.back().nb_already_explored += 1;
	      (* Exploits.back().Already_explored)[node] = true;
	      (* Exploits.back().Interp_chosen)[node] = interp_k;
	      for (int j = 0; j < int( E.Neighbors[node].size() ) ; j++){
		int vj = E.Neighbors[node][j].first; 
		//		if (  (* Exploits.back().Already_explored)[vj] )
		//		  * (* Exploits.back().Alpins)[ E.Neighbors[node][j].second ].p( ( node < vj ) ) = Perm_en_Stock[j];
		//		else
		if (!  (* Exploits.back().Already_explored)[vj] )
		  (* Exploits.back().To_be_explored).push_back(vj);
	      }
	      
	      if (verbose_global)
		cout << "    \\_ Berserk mode : new Exploration +++++++++++++++++++++++++" << endl;
	      
	    } else { // not berserk
	      v_trouve++;
	      if (verbose_global)
		cout << "    \\_ other possible interpretation" << endl;
	    }
	  } 
	  
	  //	  Perm_en_Stock.clear();
	  
	} // End loop possibilities
	
	  // if nothing found, we can kill the whole interpretation
	
	if (v_trouve == 0){
	  Dorade.empty();
	  Exploits.pop_front();
	  if (verbose_global)
	    cout << "   \\_ (This exploration falied...) Next exploration !" << endl;
	  
	  break;
	}
	
	// Otherwise, depending on the mode...
	
	if ((mode_bourrin) || (v_trouve == 1)) {
	  
	  nb_already_explored += 1;
	  Already_explored[node] = true;
	  Interp_chosen[node] = v_interp_k;
	  for (int j = 0; j < int( E.Neighbors[node].size() ) ; j++){
	    int vj = E.Neighbors[node][j].first; 
	    //	    if ( Already_explored[vj] )
	    //	      Alpins[ E.Neighbors[node][j].second ].p( ( node < vj ) ) = new vector<int>(v_Perm_en_Stock[j]);	 
	    //	    //	      * Alpins[ E.Neighbors[node][j].second ].p( ( node < vj ) ) = v_Perm_en_Stock[j];
	    //	    else
	    if (! Already_explored[vj] )
	      To_be_explored.push_back(vj);
	  }
	  
	  if (verbose_global)
	    cout << "    \\_ Updating the Exploration for the first valid interpretation (one more resolved variable)" << endl;
	  
	} else { // we're not berserk, and there are several possibilities : then we just count...
	  
	  To_be_explored_later.push_back( make_pair(node, v_trouve) );
	  
	  if (verbose_global)
	    cout << "    \\_ Failure for this variable (to many possibilities) : we try another variable." << endl;
	  
	}
	//	v_Perm_en_Stock.clear();
	
	
      } // End of an Exploration (or at least of a connected component)
      
      
      if (Dorade.To_be_explored == NULL) // try another exploration
	break;
      
      // Bit of exploration successfull
      //	if (Dorade.To_be_explored != NULL){ 
      To_be_explored_later.clear();
      if (nb_already_explored == E.nb_var) // Exploration finished
	break;
      
      // Exploration of a new connected component if necessary
      if (verbose_a_bit_more)
	cout << "Topological issue : non-connected graph : nb_already_explored = " << nb_already_explored << " vs nb_var = " << E.nb_var << endl;
      else {
	std::cout << BoldCyan << "+" << TextNormal; std::cout.flush();
      }
      int nouveau_depart = find_value(Already_explored, false);
      assert(nouveau_depart != -1);
      
      // Let's choose a variable with few possible interpretations to start again... (to reduce complexity)
      int nb_inter = Coherence[nouveau_depart].size();
      for (int i = nouveau_depart+1; i<E.nb_var; i++){
	if (Already_explored[i])
	  continue;
	int nb2 = Coherence[i].size();
	if (nb2 < nb_inter){
	  nouveau_depart = i;
	  nb_inter = nb2;
	}
	if (nb_inter == 2) // one cannot do better
	  break;
      }
      
      if (verbose_global)
	cout << endl << "Index new starting variable : x_" << nouveau_depart << ", with " << nb_inter << " interpretations." << endl;
      
      
      // Now let's generate parallel explorations
      
      for (int nk = 1; nk< int( Coherence[nouveau_depart].size() ) ; nk++){ // add each possible interpretation (except first one, considered immediatly)
	Exploits.push_back( Exploration(To_be_explored, To_be_explored_later, Already_explored, nb_already_explored, Interp_chosen, true) ); // full copy of tables
	Exploits.back().nb_already_explored += 1;
	(* Exploits.back().Already_explored)[nouveau_depart] = true;
	(* Exploits.back().Interp_chosen)[nouveau_depart] = nk;
	for (int j = 0; j < int( E.Neighbors[nouveau_depart].size() ) ; j++){
	  int vj = E.Neighbors[nouveau_depart][j].first; 
	  if (!  (* Exploits.back().Already_explored)[vj] ) // this caution should be useless
	    (* Exploits.back().To_be_explored).push_back(vj);
	}
	
	if (verbose_global)
	  cout << "    \\_ Adding a connected component : new Exploration +++++++++++++++++++++++++" << endl;
      }
      
      // Then let's update this exploration (first interpretation)
      
      nb_already_explored += 1;
      Already_explored[nouveau_depart] = true;
      Interp_chosen[nouveau_depart] = 0;
      for (int j = 0; j < int( E.Neighbors[nouveau_depart].size() ) ; j++){
	int vj = E.Neighbors[nouveau_depart][j].first;
	if (! Already_explored[vj] )
	  To_be_explored.push_back(vj);
      }
      
      if (verbose_global)
	cout << "    \\_ Updating the current Exploration for the first interpretation of the supplementary connected component." << endl;
      

	
    } while (true);// end loop connected components of a same exploration
    
    Already_explored.clear();
      
    // all fields empty except last ones
    //	Solutions.push_back( Exploration(To_be_explored, To_be_explored_later, Already_explored, 0, Interp_chosen /*, Alpins*/) );
    if (Interp_chosen.size() != 0){ // if the mission was a success
      Solutions.push_back(Interp_chosen);
      Dorade.empty();

      //	Exploits.pop_front();
      // Actually, we can empty everything, no need to seatch for another graph...
      // must empty more cleanly huh...
      
      for (list<Exploration>::iterator ex = Exploits.begin(); ex != Exploits.end(); ++ex)
	(*ex).empty();
      Exploits.clear();
    break;
    }    
    
    
  } // end of all explorations
  
  
}




/****************************************** Test (not translated) ********************************/

// This testes the code with random energies in F0

void test_alphas_random(){
  bool alpha_verbose = false;
  bool alpha2_verbose = false;
  bool methode_graph = true; // avec alphas < 0 : donne un autre resultat ; detecte les ambiguites eventuelles mais se plante generalement quand meme (forcement...)
  bool methode_histo = false; // idem pour alphas < 0                      ; plante si alphas_i ou sum_k alphas_k similaires ou 0

  double epsi = 0.000000001;
  randomness_initialization();

  int nb_bcp = 200;

  for (int iloop = 0; iloop<100000; iloop++){
    
    if (iloop%100 == 0){
      cout << "+"; cout.flush();
    }
    // int n = 4;  // nb variables (beta)
    int n = 2 + ( rand()%5 );
    int m = 2 + ( rand()%5 );
  if (alpha2_verbose)
    cout << endl << endl << "---------------------------------------" << endl << endl << "n : " << n << endl;

  /*********  Generation matrix random ***********/
  DoubleImage alphas_init = generate_alphas(n,m);  // positifs !
  //  alphas_init.nul();
  //  alphas_init(3,3) = 1;

  if (alpha_verbose){
    cout << "Alphas :"; alphas_init.display_text(5);
  }
  DoubleImage Mat_init = alphas_to_matrix(alphas_init);
  if (alpha_verbose){
    cout << "Matrix initiale: " << Mat_init.nx << " x " << Mat_init.ny << endl;
    Mat_init.display_text(5);
  }
  vector<int> permut_init_x = generate_permutation(Mat_init.nx);
  vector<int> permut_init_y = generate_permutation(Mat_init.ny);
  DoubleImage M = permute_x(Mat_init, permut_init_x);
  //  Mat_init.empty();
  M = permute_y(M, permut_init_y);
  //  M.display_text(5);


  //  cout << "Permut x : "; display_text(permut_init_x);
  //  cout << "Permut y : "; display_text(permut_init_y);



  /************   On retrouve la permut   ****************/
  pair<int,int> Imax = find_biggest(M);
  // cout << "Imax : " << Imax.first << " x " << Imax.second << " : " << M(Imax.first,Imax.second) << endl;
  //  vector<double> combines = extract_x(M, Imax.second);
  vector<double> combines_y = extract_y(M, Imax.first);
  //vector<double> combines = generate_betas_combined(n);
  //  cout << "Combines : "; display_text(combines);

  vector<double> betas_y;
  if (methode_histo)
    betas_y = find_betas_histo(combines_y);
  vector<double> betas_y2;
  if (methode_graph)
    betas_y2 = find_betas_graph(combines_y);
  double diff_by;
  if (methode_histo && methode_graph){
    diff_by = difference_max(betas_y, betas_y2);
    if (diff_by > epsi){
      cout << "Les 2 methodes de retrouvage de betas ne donnent pas le meme resultat : norme_infinie = " << diff_by << endl;
      cout << "betas 1 : "; display_text(betas_y);
      cout << "betas 2 : "; display_text(betas_y2);
    } else {
      if (alpha2_verbose)
	cout << "Les 2 methodes concordent." << endl;
    }  
  }

  //  cout << "Betas trouves: "; display_text(betas_y);
  pair< vector<int>, vector<vector<int> >   > Permut_y;
  if (methode_histo)
    Permut_y = find_permutation(betas_y, combines_y); // permutation + ambiguits
  else
    Permut_y = find_permutation(betas_y2, combines_y);

  combines_y.clear();
  DoubleImage Mp = permute_y(M, Permut_y.first);


  vector<double> combines_x = extract_x(M, Imax.second);
  vector<double> betas_x; 
  if (methode_histo)
    betas_x = find_betas_histo(combines_x);
  vector<double> betas_x2; 
  if (methode_graph)
    betas_x2 = find_betas_graph(combines_x);
  double diff_bx;
  if (methode_histo && methode_graph){
    diff_bx = difference_max(betas_x, betas_x2);
    if (diff_bx > epsi){
      cout << "Les 2 methodes de retrouvage de betas ne donnent pas le meme resultat : norme_infinie = " << diff_bx << endl;
      cout << "betas 1 : "; display_text(betas_x);
      cout << "betas 2 : "; display_text(betas_x2);
    } else {
      if (alpha2_verbose)
	cout << "Les 2 methodes concordent." << endl;
    }  
  }

  //  cout << "Betas trouves: "; display_text(betas_x);
  pair< vector<int>, vector<vector<int> >   > Permut_x;
  if (methode_histo)
    Permut_x = find_permutation(betas_x, combines_x);
  else
    Permut_x = find_permutation(betas_x2, combines_x);

  combines_x.clear();
  // attention, le deuxieme coup, on ordonne la dernire ligne et non pas la premire
  if (alpha_verbose){
    cout << "Permut x: "; display_text(Permut_x.first); 
    cout << "Ambiguites : " << Permut_x.second.size() << endl;
    cout << "Inverse : "; display_text(reverse(Permut_x.first.size()));
  }

  Permut_x.first = compose_permutations(Permut_x.first, reverse(Permut_x.first.size()));
  reverse_index(Permut_x.second, Permut_x.first.size());

  if (alpha_verbose){
    cout << "Permut x: "; display_text(Permut_x.first);
  }  
  Mp = permute_x(Mp, Permut_x.first);

  if (alpha_verbose)
    Mp.display_text(5);

  //  cout << "Permut x : "; display_text(permut_x);
  //  cout << "Permut y : "; display_text(permut_y);


  /*************  puis les alphas    ************/
  DoubleImage alpha( (methode_histo)?(betas_x.size()):(betas_x2.size()), 
		     (methode_histo)?(betas_y.size()):(betas_y2.size()) );
  if (methode_histo){
    betas_x.clear();
    betas_y.clear();
  }
  if (methode_graph){
    betas_y2.clear();
    betas_x2.clear();
  }

  /************ rordonnancements possibles, comparaison entre alphas *********/
  DoubleImage alphas_init_ord = sort_decreasing_sums(alphas_init);
  if (alpha_verbose){
    cout << "Alphas init reord: "; alphas_init_ord.display_text(5);
  }
  //  cout << "Alphas init      : "; alphas_init.display_text(5);
  
  DoubleImage alpha_rmb(alpha.nx, alpha.ny); // remember my old good alpha...

  
  for (int i = 0; i<alpha.nx; i++)
    for (int j = 0; j<alpha.ny; j++)
      alpha(i,j) = matrix_to_alpha(Mp, i, j);
  
  double diff_min = difference_max(alphas_init_ord, alpha);
  int nbpt = 1; // nb de permutations testes (id = 1...)
  int nbpp = 0; // nb de matrices valables (appartenant  la family aprs permutation)
  if (is_in_the_family(Mp,alpha))
    nbpp++;

  if  (   ( Permut_x.second.size() != 0 ) ||  ( Permut_y.second.size() != 0 )    ){ // ambiguits : grer toutes les permut'

    //    cout << "(iloop = " << iloop << ")   Matrix " <<  Mp.nx << "x" <<  Mp.ny << ".  Ambiguites : " << Permut_x.second.size() << " x " << Permut_y.second.size() << endl;
    //    cout << "Mat retrouvee : "; Mp.display_text();

    for (int i = 0; i<alpha.nx; i++)
      for (int j = 0; j<alpha.ny; j++)
	alpha_rmb(i,j) = alpha(i, j);

    int np_tot = Permut_x.second.size() + Permut_y.second.size();
    int np_1 = Permut_x.second.size();
    vector<int>  coord(np_tot, 0);     // tableau-coordonnes d'une permutation
    //    vector< vector<int> > Permapp(np_tot); // les permutations successives  applyr
    vector<bool> length_cycles(max(Mp.nx, Mp.ny), false); // on liste les permutations gnriques  prparer
    int nb_futur = 1;
    for (int k = 0; k<np_1; k++){
      length_cycles[ Permut_x.second[k].size() ] = true;
      nb_futur *=  factorielle(Permut_x.second[k].size());
    }
    for (int k = 0; k<int(Permut_y.second.size()); k++){
      length_cycles[ Permut_y.second[k].size() ] = true;
      nb_futur *=  factorielle(Permut_y.second[k].size());
    }
    if (nb_futur > nb_bcp){
      cout << "*" << nb_futur; cout.flush();
    }

    vector< vector< vector<int> > > Perm_pures = generate_all_permutations(length_cycles);
    
    for (int k = 0; k<int(Perm_pures.size()); k++)
      if (length_cycles[k])
	modified_sequence_of_permutations(Perm_pures[k]);
    
    // Plutot a faire directement sur la matrix (plus rapide si elle est grosse)
    //    for (int k = 0; k<np_1; k++) // [1 3 2](5, 12, 53) = [5, 53, 12]
    //      Permapp[k] = remplace_indexes( Perm_pures[ Permut_x.second[k].size() ],  Permut_x.second[k] )
    //    for (int k = np_1; k<np_tot; k++)
    //      Permapp[k] = remplace_indexes( Perm_pures[ Permut_y.second[k-np_1].size() ],  Permut_y.second[k-np_1] )
    

    do {

      int deg = 0;
      bool cran_de_plus;
      do {
	cran_de_plus = false;

	if (deg<np_1){
	  apply_permut_index_x(Mp, Permut_x.second[deg], Perm_pures[ Permut_x.second[deg].size() ][ coord[deg] ] );  // matrix, indexes, permut
	  //	  cout << "x" << endl;
	  //	  cout << "Permu: "; display_text(Perm_pures[ Permut_x.second[deg].size() ][ coord[deg] ]);
	  //	  cout << "Index: "; display_text( Permut_x.second[deg]);
	  coord[deg]++;
	  //	  if (coord[deg] ==  int( Permut_x.second[deg].size() )  ){
	  if (coord[deg] ==  int( Perm_pures[ Permut_x.second[deg].size() ].size() )  ){
	    cran_de_plus = true;
	    coord[deg] = 0;
	    deg++;
	  }
	} else {
	  apply_permut_index_y(Mp, Permut_y.second[deg-np_1], Perm_pures[ Permut_y.second[deg-np_1].size() ][ coord[deg] ] );  // matrix, indexes, permut
	  //	  cout << "y" << endl;
	  //	  cout << "Permu: "; display_text(Perm_pures[ Permut_y.second[deg-np_1].size() ][ coord[deg] ]);
	  //	  cout << "Index: "; display_text( Permut_y.second[deg-np_1]);
	  coord[deg]++;
	  //	  if (coord[deg] ==  int( Permut_y.second[deg-np_1].size() ) ){
	  if (coord[deg] ==  int( Perm_pures[ Permut_y.second[deg-np_1].size() ].size() ) ){
	    cran_de_plus = true;
	    coord[deg] = 0;
	    deg++;
	  }
	}

	//cout << "Coord: "; display_text(coord);
	//	cout << "Mat retrouvee : "; Mp.display_text();
	//	string s; getline(cin, s);
      }
      while (cran_de_plus && (deg<np_tot));

      //      cout << "Une permut' finie !!!!!!!!!!!!!!" << endl;
      if (cran_de_plus) // toutes permut' testes, retour  l'tat initial
	break;



      for (int i = 0; i<alpha.nx; i++)
	for (int j = 0; j<alpha.ny; j++)
	  alpha(i,j) = matrix_to_alpha(Mp, i, j);
      
      nbpt++;
      if (is_in_the_family(Mp,alpha))
	nbpp++;

      double diff_nouv = difference_max(alphas_init_ord, alpha);
      if (diff_nouv < diff_min){
	diff_min = diff_nouv;
	for (int i = 0; i<alpha.nx; i++)
	  for (int j = 0; j<alpha.ny; j++)
	    alpha_rmb(i,j) = alpha(i, j);

	if (diff_min < epsi) // a sert  rien de continuer  chercher quand on a trouv la solution
	  break;
      }

      
    } while (true);


    for (int i = 0; i<alpha.nx; i++)
      for (int j = 0; j<alpha.ny; j++)
	alpha(i,j) = alpha_rmb(i, j);
    alpha_rmb.empty();

    if (nb_futur > nb_bcp){
      cout << "."; cout.flush();
    }
  }
  
  
  
  if (alpha_verbose){
    cout << "Alpha retrouves: "; alpha.display_text(5);
  }

  if ((nbpt != 1) || (nbpp != nbpt))
  cout << "| " << nbpp << "/" << nbpt; cout.flush();

  if (diff_min > epsi){
    cout << endl << "   ----------------------------------" << endl << endl;
    cout << "(iloop = " << iloop << ")  ERREUR !!! Difference : " << difference_max(alphas_init_ord, alpha) << endl;
    cout << "Nb perm testees: " << nbpt << endl;
    cout << "(iloop = " << iloop << ")   Matrix " <<  Mp.nx << "x" <<  Mp.ny << ".  Ambiguites : " << Permut_x.second.size() << " x " << Permut_y.second.size() << endl;
    cout << "Mat init : "; Mat_init.display_text();
    cout << "Mat trouvee : "; Mp.display_text();
  }
  Mat_init.empty();

  alpha.empty();
  alphas_init_ord.empty();
  alphas_init.empty();
  M.empty();
  Mp.empty();

  }  // fin boucle sur les tests alatoires


}



} // end of namespace

#endif

